from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.credential_base import CredentialBaseAttributes


class CertificateCredentialAttributes(CredentialBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Certificate Credential"
	certificate = Attribute('Certificate')



from pytpp.properties.response_objects.dataclasses import stats
from pytpp.tools.helpers.date_converter import from_date_string


class Stats:
    @staticmethod
    def Counter(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return stats.Counter(
            a_name=response_object.get('AName'),
            b_name=response_object.get('BName'),
            c_name=response_object.get('CName'),
            description=response_object.get('Description'),
            name=response_object.get('Name'),
            stats_type=response_object.get('StatsType'),
        )

    @staticmethod
    def Result(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return stats.Result(
            key=Stats.Key(response_object.get('Key')),
            value=[Stats.Value(value) for value in response_object.get('Value')],
        )

    @staticmethod
    def Key(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return stats.Key(
            m_item1=response_object.get('m_Item1'),
            m_item2=response_object.get('m_Item2'),
            m_item3=response_object.get('m_Item3'),
        )

    @staticmethod
    def Value(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return stats.Value(
            count=response_object.get('Count', 0),
            sum_value=response_object.get('SumValue', 0),
            tag_a=response_object.get('TagA'),
            tag_b=response_object.get('TagB'),
            tag_c=response_object.get('TagC'),
            time_frame=from_date_string(response_object.get('TimeFrame')),
            type=response_object.get('Type'),
        )



from pytpp.features.bases.feature_base import FeatureBase, feature
from pytpp.features.definitions.exceptions import InvalidResultCode
from pytpp.properties.config import ClientGroupsAttributeValues
from pytpp.attributes.client_group import ClientGroupAttributes
from typing import Union, TYPE_CHECKING
if TYPE_CHECKING:
    from pytpp.tools.vtypes import Config


class _ClientGroupBase(FeatureBase):
    def __init__(self, api):
        super().__init__(api)

        self._group_base_dn = r'\VED\Clients\Groups'
        self._work_base_dn = r'\VED\Clients\Work'

    def assign_work(self, group: 'Union[Config.Object, str]', work: 'Union[Config.Object, str]'):
        """
        Assigns work to the client group

        Args:
            group: The :ref:`config_object` or name of the client group.
            work: The :ref:`config_object` or name of the work.
        """
        group_dn = self._get_dn(group, parent_dn=self._group_base_dn)
        work_dn = self._get_dn(work, parent_dn=self._work_base_dn)
        response = self._api.websdk.Config.Write.post(
            object_dn=group_dn,
            attribute_data=self._name_value_list({
                ClientGroupAttributes.assigned_work: [work_dn]
            })
        )

        if response.result.code != 1:
            raise InvalidResultCode(code=response.result.code,
                                                 code_description=response.result.credential_result)

    def delete(self, group: 'Union[Config.Object, str]'):
        """
        Deletes a client group

        Args:
            group: :ref:`config_object` or name of the client group.
        """
        group_dn = self._get_dn(group, parent_dn=self._group_base_dn)
        self._config_delete(object_dn=group_dn)

    def get(self, name: str, raise_error_if_not_exists: bool = True):
        """
        Args:
            name: The name of the client group.
            raise_error_if_not_exists: Raise an exception if the client group does not exist.

        Returns:
            :ref:`config_object` of the client group.
        """
        return self._get_config_object(
            object_dn=fr'{self._group_base_dn}\{name}',
            raise_error_if_not_exists=raise_error_if_not_exists
        )

    def list(self):
        """
        Lists all client groups.

        Returns:
            List of :ref:`config_object` of the client groups.
        """
        response = self._api.websdk.Config.Enumerate.post(object_dn=self._group_base_dn)

        if response.result.code != 1:
            raise InvalidResultCode(
                code=response.result.code,
                code_description=response.result.credential_result
            )
        return response.objects

    def remove_work(self, group: 'Union[Config.Object, str]', work: 'Union[Config.Object, str]'):
        """
        Removes work from a client group

        Args:
            group: The :ref:`config_object` or name of the client group.
            work: The :ref:`config_object` or name of the work to be removed.
        """
        group_dn = self._get_dn(group, parent_dn=self._group_base_dn)
        work_dn = self._get_dn(work, parent_dn=self._work_base_dn)
        response = self._api.websdk.Config.RemoveDnValue.post(
            object_dn=group_dn,
            attribute_name=ClientGroupAttributes.assigned_work,
            value=work_dn
        )
        if response.result.code != 1:
            raise InvalidResultCode(
                code=response.result.code,
                code_description=response.result.credential_result
            )


@feature('Agentless Group')
class Agentless(_ClientGroupBase):
    def create(self, name: str, get_if_already_exists: bool = True):
        """
        Args:
            name: The name of the client group.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the client group.
        """
        attributes = {
            ClientGroupAttributes.created_by: ClientGroupsAttributeValues.CreatedBy.websdk,
            ClientGroupAttributes.agent_type: ClientGroupsAttributeValues.AgentType.agentless,
            ClientGroupAttributes.rule      : ClientGroupsAttributeValues.DefaultRules.agentless
        }

        return self._config_create(
            name=name,
            parent_folder_dn=self._group_base_dn,
            config_class=ClientGroupAttributes.__config_class__,
            attributes=attributes,
            get_if_already_exists=get_if_already_exists
        )


@feature('EST Certificate Enrollment Group')
class EstCertificateEnrollment(_ClientGroupBase):
    def create(self, name: str, get_if_already_exists: bool = True):
        """
        Args:
            name: The name of the client group.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the client group.
        """
        attributes = {
            ClientGroupAttributes.created_by: ClientGroupsAttributeValues.CreatedBy.websdk,
            ClientGroupAttributes.agent_type: ClientGroupsAttributeValues.AgentType.est,
            ClientGroupAttributes.rule      : ClientGroupsAttributeValues.DefaultRules.est
        }

        return self._config_create(
            name=name,
            parent_folder_dn=self._group_base_dn,
            config_class=ClientGroupAttributes.__config_class__,
            attributes=attributes,
            get_if_already_exists=get_if_already_exists
        )


@feature('Venafi Agent Group')
class VenafiAgent(_ClientGroupBase):
    def create(self, name: str, get_if_already_exists: bool = True):
        """
        Args:
            name: The name of the client group.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            :ref:`config_object` of the client group.
        """
        attributes = {
            ClientGroupAttributes.created_by: ClientGroupsAttributeValues.CreatedBy.websdk,
            ClientGroupAttributes.agent_type: ClientGroupsAttributeValues.AgentType.venafi_agent,
            ClientGroupAttributes.rule      : ClientGroupsAttributeValues.DefaultRules.venafi_agent
        }

        return self._config_create(
            name=name,
            parent_folder_dn=self._group_base_dn,
            config_class=ClientGroupAttributes.__config_class__,
            attributes=attributes,
            get_if_already_exists=get_if_already_exists
        )




from pytpp.attributes._helper import IterableMeta, Attribute


class SshDeviceBaseAttributes(metaclass=IterableMeta):
	__config_class__ = "Ssh Device Base"
	algorithm = Attribute('Algorithm', min_version='15.1')
	allow_duplicate_private_keys = Attribute('Allow Duplicate Private Keys', min_version='15.1')
	allow_from = Attribute('Allow From', min_version='15.1')
	allow_multiple_authorized_keys = Attribute('Allow Multiple Authorized Keys', min_version='15.1')
	allow_root_access = Attribute('Allow Root Access', min_version='15.1')
	allow_shared_server_accounts = Attribute('Allow Shared Server Accounts', min_version='15.1')
	allow_ssh1 = Attribute('Allow Ssh1', min_version='15.1')
	allow_unencrypted_private_keys = Attribute('Allow Unencrypted Private Keys', min_version='15.1')
	allowed_algorithm = Attribute('Allowed Algorithm', min_version='15.1')
	allowed_command = Attribute('Allowed Command', min_version='15.1')
	allowed_vendor_types = Attribute('Allowed Vendor Types', min_version='15.1')
	automatic_rotation_cleanup_wait = Attribute('Automatic Rotation Cleanup Wait', min_version='15.1')
	automatic_rotation_enabled = Attribute('Automatic Rotation Enabled', min_version='15.1')
	automatic_rotation_interval = Attribute('Automatic Rotation Interval', min_version='15.1')
	automatic_rotation_lead_time = Attribute('Automatic Rotation Lead Time', min_version='15.1')
	deny_from = Attribute('Deny From', min_version='15.1')
	environment = Attribute('Environment', min_version='19.2')
	host_trusts = Attribute('Host Trusts', min_version='15.1')
	key_bit_strength = Attribute('Key Bit Strength', min_version='15.1')
	known_hosts = Attribute('Known Hosts', min_version='15.1')
	management_type = Attribute('Management Type', min_version='15.1')
	maximum_key_age = Attribute('Maximum Key Age', min_version='15.1')
	minimum_key_bit_strength = Attribute('Minimum Key Bit Strength', min_version='15.1')
	required_options = Attribute('Required Options', min_version='15.1')
	root_server_access = Attribute('Root Server Access', min_version='15.1')
	server_access = Attribute('Server Access', min_version='15.1')
	ssh_device_status = Attribute('Ssh Device Status', min_version='15.1')
	ssh_device_type = Attribute('Ssh Device Type', min_version='15.1')
	trusted_root_users = Attribute('Trusted Root Users', min_version='15.1')
	trusted_users = Attribute('Trusted Users', min_version='15.1')
	update_cache = Attribute('Update Cache', min_version='15.1')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_base import ApplicationBaseAttributes


class AmazonAppAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Amazon App"
	access_key_id = Attribute('Access Key ID', min_version='16.1')
	aws_credential_dn = Attribute('Aws Credential DN', min_version='18.3')
	binding_target = Attribute('Binding Target', min_version='16.1')
	certificate_arn = Attribute('Certificate ARN', min_version='16.1')
	certificate_name = Attribute('Certificate Name', min_version='16.1')
	cloudfront_distribution_id = Attribute('CloudFront Distribution ID', min_version='16.1')
	create_binding = Attribute('Create Binding', min_version='18.2')
	file_validation_disabled = Attribute('File Validation Disabled', min_version='16.1')
	iam_certificate_id = Attribute('IAM Certificate ID', min_version='16.1')
	initial_binding_attempt = Attribute('Initial Binding Attempt', min_version='16.1')
	install_path = Attribute('Install Path', min_version='16.1')
	issued_by_aws = Attribute('Issued By AWS', min_version='16.1')
	load_balancer_name = Attribute('Load Balancer Name', min_version='16.1')
	load_balancer_port = Attribute('Load Balancer Port', min_version='16.1')
	load_balancer_region_code = Attribute('Load Balancer Region Code', min_version='16.1')
	minimum_protocol_version = Attribute('Minimum Protocol Version', min_version='21.3')
	network_validation_disabled = Attribute('Network Validation Disabled', min_version='16.1')
	provisioning_to = Attribute('Provisioning To', min_version='18.2')
	region_code = Attribute('Region Code', min_version='16.1')
	replace_store = Attribute('Replace Store', min_version='18.3')
	secret_access_key = Attribute('Secret Access Key', min_version='16.1')
	target_group = Attribute('Target Group', min_version='18.2')
	timeout = Attribute('Timeout', min_version='16.1')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.log_channel import LogChannelAttributes


class LogAdaptableAttributes(LogChannelAttributes, metaclass=IterableMeta):
	__config_class__ = "Log Adaptable"
	adaptable_logging_channel_text_field_1 = Attribute('Adaptable Logging Channel Text Field 1', min_version='17.1')
	adaptable_logging_channel_text_field_10 = Attribute('Adaptable Logging Channel Text Field 10', min_version='17.1')
	adaptable_logging_channel_text_field_11 = Attribute('Adaptable Logging Channel Text Field 11', min_version='17.1')
	adaptable_logging_channel_text_field_12 = Attribute('Adaptable Logging Channel Text Field 12', min_version='17.1')
	adaptable_logging_channel_text_field_2 = Attribute('Adaptable Logging Channel Text Field 2', min_version='17.1')
	adaptable_logging_channel_text_field_3 = Attribute('Adaptable Logging Channel Text Field 3', min_version='17.1')
	adaptable_logging_channel_text_field_4 = Attribute('Adaptable Logging Channel Text Field 4', min_version='17.1')
	adaptable_logging_channel_text_field_5 = Attribute('Adaptable Logging Channel Text Field 5', min_version='17.1')
	adaptable_logging_channel_text_field_6 = Attribute('Adaptable Logging Channel Text Field 6', min_version='17.1')
	adaptable_logging_channel_text_field_7 = Attribute('Adaptable Logging Channel Text Field 7', min_version='17.1')
	adaptable_logging_channel_text_field_8 = Attribute('Adaptable Logging Channel Text Field 8', min_version='17.1')
	adaptable_logging_channel_text_field_9 = Attribute('Adaptable Logging Channel Text Field 9', min_version='17.1')
	credential = Attribute('Credential', min_version='17.1')
	log_debug = Attribute('Log Debug', min_version='19.3')
	oauth_token_application_id = Attribute('OAuth Token Application Id', min_version='21.3')
	oauth_token_credential = Attribute('OAuth Token Credential', min_version='21.3')
	oauth_token_scope = Attribute('OAuth Token Scope', min_version='21.3')
	powershell_script = Attribute('PowerShell Script', min_version='17.1')
	powershell_script_hash_vault_id = Attribute('PowerShell Script Hash Vault Id', min_version='18.4')
	script_execution_timeout = Attribute('Script Execution Timeout', min_version='20.2')
	secondary_credential = Attribute('Secondary Credential', min_version='17.4')
	service_address = Attribute('Service Address', min_version='17.1')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.http_ca_base import HTTPCABaseAttributes


class GoogleCloudCAAttributes(HTTPCABaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Google Cloud CA"
	certificate_authority_id = Attribute('Certificate Authority ID', min_version='20.2')
	region = Attribute('Region', min_version='20.4')



from pytpp.tools.deprecation import DeprecationMeta
from pytpp.properties.config import (
    FolderAttributes, ClientGroupsAttributes, ClientWorkAttributes, WorkflowAttributes, CustomFieldAttributes,
    CertificateAttributes, DeviceAttributes, ApplicationAttributes, CredentialAttributes, DiscoveryAttributes,
    CertificateAuthorityAttributes, PlatformsAttributes, PlacementRulesAttributeNames, ApplicationGroupAttributes
)

class AttributeNames(metaclass=DeprecationMeta):
    __deprecation_reason__ = 'Using AttributeNames will be deprecated soon. Please use ' \
                             'pytpp.Attributes (from pytpp import Attributes) instead.'
    Application = ApplicationAttributes
    ApplicationGroup = ApplicationGroupAttributes
    Certificate = CertificateAttributes
    CertificateAuthority = CertificateAuthorityAttributes
    ClientGroups = ClientGroupsAttributes
    ClientWork = ClientWorkAttributes
    Credentials = CredentialAttributes
    CustomField = CustomFieldAttributes
    Device = DeviceAttributes
    Discovery = DiscoveryAttributes
    Folder = FolderAttributes
    PlacementRules = PlacementRulesAttributeNames
    Platforms = PlatformsAttributes
    Workflow = WorkflowAttributes



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class F5AuthenticationBundleAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "F5 Authentication Bundle"
	advanced_settings_bundle_name = Attribute('Advanced Settings Bundle Name')
	certificates = Attribute('Certificates')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.http_ca_base import HTTPCABaseAttributes


class KeynectisSequoiaCAAttributes(HTTPCABaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Keynectis Sequoia CA"
	csr_format_name = Attribute('CSR Format Name')
	csr_name = Attribute('CSR Name')
	customer = Attribute('Customer')
	encryption_certificate = Attribute('Encryption Certificate')
	enrollment_mode = Attribute('Enrollment Mode')
	enterprise = Attribute('Enterprise')
	fields = Attribute('Fields')
	offer = Attribute('Offer')
	uri = Attribute('URI')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class FlowGroupAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Flow Group"
	product_code_description = Attribute('Product Code Description', min_version='19.4')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.key_base import KeyBaseAttributes


class SecretKeyAttributes(KeyBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Secret Key"
	secret_key_renewal_flow = Attribute('Secret Key Renewal Flow', min_version='20.2')



from pytpp.attributes._helper import IterableMeta
from pytpp.attributes.client_group_base import ClientGroupBaseAttributes


class ClientSubgroupAttributes(ClientGroupBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Client Subgroup"



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_base import ApplicationBaseAttributes


class BlueCoatSSLVAAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "BlueCoat SSLVA"
	certificate_label = Attribute('Certificate Label')
	certificate_oid = Attribute('Certificate OID')
	certificate_only = Attribute('Certificate Only')
	create_lists = Attribute('Create Lists')
	device_certificate = Attribute('Device Certificate')
	file_validation_disabled = Attribute('File Validation Disabled')
	network_validation_disabled = Attribute('Network Validation Disabled')
	replace_store = Attribute('Replace Store')



from pytpp.attributes._helper import IterableMeta
from pytpp.attributes.service_module import ServiceModuleAttributes


class AutoLayoutManagerAttributes(ServiceModuleAttributes, metaclass=IterableMeta):
	__config_class__ = "Auto Layout Manager"



from pytpp.attributes._helper import IterableMeta

class Classes(metaclass=IterableMeta):
	a10_ax_traffic_manager = "A10 AX Traffic Manager"
	acme_account = "ACME Account"
	acme_authorization = "ACME Authorization"
	acme_challenge = "ACME Challenge"
	acme_order = "ACME Order"
	acme_root = "ACME Root"
	aws_ec2_instance_monitor = "AWS EC2 Instance Monitor"
	activedirectory_identity_driver = "ActiveDirectory Identity Driver"
	adaptable_app = "Adaptable App"
	adaptable_bulk_app = "Adaptable Bulk App"
	adaptable_ca = "Adaptable CA"
	adaptable_credential = "Adaptable Credential"
	adaptable_credential_connector = "Adaptable Credential Connector"
	adaptable_flow_action_base = "Adaptable Flow Action Base"
	adaptable_workflow = "Adaptable Workflow"
	agent = "Agent"
	agent_base = "Agent Base"
	agent_certificate_base = "Agent Certificate Base"
	agent_certificate_discovery = "Agent Certificate Discovery"
	agent_certificate_driver = "Agent Certificate Driver"
	agent_container = "Agent Container"
	agent_discovery = "Agent Discovery"
	agent_discovery_base = "Agent Discovery Base"
	agent_driver = "Agent Driver"
	agent_driver_base = "Agent Driver Base"
	agent_module_handler = "Agent Module Handler"
	agent_module_handler_container = "Agent Module Handler Container"
	agent_root = "Agent Root"
	agent_ssh_base = "Agent SSH Base"
	agent_ssh_discovery = "Agent SSH Discovery"
	agent_ssh_driver = "Agent SSH Driver"
	amazon_app = "Amazon App"
	amazon_ca = "Amazon CA"
	amazon_credential = "Amazon Credential"
	apache = "Apache"
	apache_application_group = "Apache Application Group"
	aperture_configuration = "Aperture Configuration"
	application_base = "Application Base"
	application_group = "Application Group"
	associate_key_expiration_action = "Associate Key Expiration Action"
	auto_layout_manager = "Auto Layout Manager"
	automatic_password_credential = "Automatic Password Credential"
	azure_key_vault = "Azure Key Vault"
	basic = "Basic"
	bluecoat_sslva = "BlueCoat SSLVA"
	bluecoat_sslva_trust_store = "BlueCoat SSLVA Trust Store"
	branch_base = "Branch Base"
	brocade = "Brocade"
	bulk_application_base = "Bulk Application Base"
	bulk_provisioning_manager = "Bulk Provisioning Manager"
	ca_container = "CA Container"
	ca_import = "CA Import"
	ca_import_manager = "CA Import Manager"
	ca_root = "CA Root"
	capi = "CAPI"
	capi_trust_store = "CAPI Trust Store"
	csp_key_container = "CSP Key Container"
	certificate_authority_base = "Certificate Authority Base"
	certificate_credential = "Certificate Credential"
	certificate_manager = "Certificate Manager"
	certificate_pre_enrollment = "Certificate Pre-Enrollment"
	certificate_provisioning_work = "Certificate Provisioning Work"
	certificate_revocation = "Certificate Revocation"
	certificate_trust_bundle = "Certificate Trust Bundle"
	certificate_trust_store_base = "Certificate Trust Store Base"
	ciscoace = "CiscoACE"
	ciscocsm = "CiscoCSM"
	ciscocss = "CiscoCSS"
	client_agent_automatic_upgrade_work = "Client Agent Automatic Upgrade Work"
	client_agent_configuration_work = "Client Agent Configuration Work"
	client_agent_device_placement_work = "Client Agent Device Placement Work"
	client_agent_ssh_discovery_work = "Client Agent SSH Discovery Work"
	client_agent_ssh_key_usage_work = "Client Agent SSH Key Usage Work"
	client_agent_ssh_provisioning_work = "Client Agent SSH Provisioning Work"
	client_certificate_discovery_work = "Client Certificate Discovery Work"
	client_certificate_work = "Client Certificate Work"
	client_group = "Client Group"
	client_group_base = "Client Group Base"
	client_group_container = "Client Group Container"
	client_group_root = "Client Group Root"
	client_portal_base = "Client Portal Base"
	client_root = "Client Root"
	client_subgroup = "Client Subgroup"
	client_user_certificate_work = "Client User Certificate Work"
	client_work_base = "Client Work Base"
	client_work_container = "Client Work Container"
	client_work_root = "Client Work Root"
	clientgroup_monitor = "ClientGroup Monitor"
	cloud_instance_monitor = "Cloud Instance Monitor"
	cloud_instance_monitor_driver_base = "Cloud Instance Monitor Driver Base"
	code_sign_preapproval_action = "Code Sign PreApproval Action"
	code_signing_action = "Code Signing Action"
	code_signing_admin_approval_action = "Code Signing Admin Approval Action"
	code_signing_apple_environment = "Code Signing Apple Environment"
	code_signing_apple_environment_template = "Code Signing Apple Environment Template"
	code_signing_application = "Code Signing Application"
	code_signing_application_collection = "Code Signing Application Collection"
	code_signing_application_container = "Code Signing Application Container"
	code_signing_application_root = "Code Signing Application Root"
	code_signing_csp_environment = "Code Signing CSP Environment"
	code_signing_csp_environment_template = "Code Signing CSP Environment Template"
	code_signing_certificate_environment = "Code Signing Certificate Environment"
	code_signing_certificate_environment_template = "Code Signing Certificate Environment Template"
	code_signing_delete_object_action = "Code Signing Delete Object Action"
	code_signing_dotnet_environment = "Code Signing DotNet Environment"
	code_signing_dotnet_environment_template = "Code Signing DotNet Environment Template"
	code_signing_environment_base = "Code Signing Environment Base"
	code_signing_environment_container = "Code Signing Environment Container"
	code_signing_environment_root = "Code Signing Environment Root"
	code_signing_environment_template_base = "Code Signing Environment Template Base"
	code_signing_environment_template_update_action = "Code Signing Environment Template Update Action"
	code_signing_flow_container = "Code Signing Flow Container"
	code_signing_flow_root = "Code Signing Flow Root"
	code_signing_gpg_environment = "Code Signing GPG Environment"
	code_signing_gpg_environment_template = "Code Signing GPG Environment Template"
	code_signing_key_pair_environment = "Code Signing Key Pair Environment"
	code_signing_key_pair_environment_template = "Code Signing Key Pair Environment Template"
	code_signing_key_time_constraint = "Code Signing Key Time Constraint"
	code_signing_pre_qualified_action = "Code Signing Pre Qualified Action"
	code_signing_project = "Code Signing Project"
	code_signing_project_container = "Code Signing Project Container"
	code_signing_project_root = "Code Signing Project Root"
	code_signing_root = "Code Signing Root"
	code_signing_time_constraint_root = "Code Signing Time Constraint Root"
	comodo_ca = "Comodo CA"
	comodo_ccm = "Comodo CCM"
	connectdirect = "ConnectDirect"
	connectdirect_trust_store = "ConnectDirect Trust Store"
	connection_base = "Connection Base"
	create_key_action = "Create Key Action"
	credential_base = "Credential Base"
	credential_container = "Credential Container"
	credential_driver_base = "Credential Driver Base"
	credential_root = "Credential Root"
	cyberark_credentials_driver = "CyberArk Credentials Driver"
	cyberark_password_credential = "CyberArk Password Credential"
	cyberark_username_password_credential = "CyberArk Username Password Credential"
	dpapi_encryption_driver = "DPAPI Encryption Driver"
	datapower = "DataPower"
	datapower_trust_store = "DataPower Trust Store"
	device = "Device"
	digicert_ca = "DigiCert CA"
	discovery = "Discovery"
	discovery_container = "Discovery Container"
	discovery_exclusion = "Discovery Exclusion"
	discovery_manager = "Discovery Manager"
	discovery_root = "Discovery Root"
	discovery_statistics = "Discovery Statistics"
	driver_base = "Driver Base"
	encryption_driver = "Encryption Driver"
	encryption_root = "Encryption Root"
	engine_root = "Engine Root"
	entrust_pki_gateway = "Entrust PKI Gateway"
	entrust_security_manager_ca = "Entrust Security Manager CA"
	entrustnet_ca = "EntrustNET CA"
	exclusion = "Exclusion"
	f5 = "F5"
	f5_authentication_bundle = "F5 Authentication Bundle"
	f5_ltm_advanced = "F5 LTM Advanced"
	f5_ltm_advanced_trust_store = "F5 LTM Advanced Trust Store"
	flow = "Flow"
	flow_action_add_team_member = "Flow Action Add Team Member"
	flow_action_approver_base = "Flow Action Approver Base"
	flow_action_base = "Flow Action Base"
	flow_action_config_read_approvers = "Flow Action Config Read Approvers"
	flow_action_direct_approvers = "Flow Action Direct Approvers"
	flow_action_team_owner_approval = "Flow Action Team Owner Approval"
	flow_event_action = "Flow Event Action"
	flow_group = "Flow Group"
	flow_root = "Flow Root"
	gsk = "GSK"
	gsk_trust_store = "GSK Trust Store"
	generational_credential = "Generational Credential"
	generic_credential = "Generic Credential"
	geotrust_ca = "Geotrust CA"
	geotrust_enterprise_ca = "Geotrust Enterprise CA"
	geotrusttrueflex_ca = "GeotrustTrueFlex CA"
	globalsign_mssl_ca = "GlobalSign MSSL CA"
	google_cloud_app = "Google Cloud App"
	google_cloud_ca = "Google Cloud CA"
	google_credential = "Google Credential"
	group = "Group"
	http_ca_base = "HTTP CA Base"
	hashicorp_vault_pki = "HashiCorp Vault PKI"
	hydrantid_ca = "HydrantId CA"
	iis5 = "IIS5"
	iis6 = "IIS6"
	identity_driver = "Identity Driver"
	identity_root = "Identity Root"
	imperva_mx = "Imperva MX"
	integrated_credential_driver = "Integrated Credential Driver"
	intermediate_and_root_certificates = "Intermediate and Root Certificates"
	jks = "JKS"
	jks_trust_store = "JKS Trust Store"
	jump_server = "Jump Server"
	juniper_sas = "Juniper SAS"
	key_base = "Key Base"
	key_manager = "Key Manager"
	key_pair = "Key Pair"
	keynectis_sequoia_ca = "Keynectis Sequoia CA"
	kmip_service_module = "Kmip Service Module"
	ldap_identity_driver = "LDAP Identity Driver"
	layer_7_ssg = "Layer 7 SSG"
	layout_driver_base = "Layout Driver Base"
	layout_driver_container = "Layout Driver Container"
	layout_root = "Layout Root"
	layout_rule_base = "Layout Rule Base"
	layout_rule_container = "Layout Rule Container"
	legacy_key_base = "Legacy Key Base"
	license = "License"
	license_root = "License Root"
	log_adaptable = "Log Adaptable"
	log_application = "Log Application"
	log_application_container = "Log Application Container"
	log_application_customization = "Log Application Customization"
	log_channel = "Log Channel"
	log_channel_container = "Log Channel Container"
	log_file = "Log File"
	log_filter = "Log Filter"
	log_heartbeat = "Log Heartbeat"
	log_mssql = "Log MSSQL"
	log_notification = "Log Notification"
	log_notification_container = "Log Notification Container"
	log_root = "Log Root"
	log_smtp = "Log SMTP"
	log_snmp = "Log SNMP"
	log_sql_channel = "Log SQL Channel"
	log_server = "Log Server"
	log_splunk = "Log Splunk"
	log_syslog = "Log Syslog"
	metadata_base = "Metadata Base"
	metadata_category = "Metadata Category"
	metadata_choice = "Metadata Choice"
	metadata_container = "Metadata Container"
	metadata_datetime = "Metadata DateTime"
	metadata_identity = "Metadata Identity"
	metadata_list = "Metadata List"
	metadata_root = "Metadata Root"
	metadata_text = "Metadata Text"
	microsoft_ca = "Microsoft CA"
	microsoft_ca_pool = "Microsoft CA Pool"
	monitoring_base = "Monitoring Base"
	monitoring_module = "Monitoring Module"
	netscaler = "NetScaler"
	network_device_certificate_work = "Network Device Certificate Work"
	null_encryption_driver = "Null Encryption Driver"
	onboard_discovery = "Onboard Discovery"
	onboard_discovery_manager = "Onboard Discovery Manager"
	openssl_ca = "OpenSSL CA"
	opentrust_pki_ca = "OpenTrust PKI CA"
	organization = "Organization"
	pem = "PEM"
	pem_trust_store = "PEM Trust Store"
	pkcs_12 = "PKCS#12"
	pkcs_12_trust_store = "PKCS#12 Trust Store"
	pkcs11 = "PKCS11"
	pkcs11_application_group = "PKCS11 Application Group"
	palo_alto_network_fw = "Palo Alto Network FW"
	palo_alto_network_fw_trust_store = "Palo Alto Network FW Trust Store"
	password_credential = "Password Credential"
	person = "Person"
	pkcs11_encryption_driver = "Pkcs11 Encryption Driver"
	placement_job = "Placement Job"
	placement_job_container = "Placement Job Container"
	policy = "Policy"
	private_key_credential = "Private Key Credential"
	proxy = "Proxy"
	proxy_base = "Proxy Base"
	quovadis_ca = "QuoVadis CA"
	rsa_keon_ca = "RSA Keon CA"
	redhat_ca = "Redhat CA"
	remote_access_application = "Remote Access Application"
	remote_access_application_container = "Remote Access Application Container"
	remote_access_root = "Remote Access Root"
	remote_access_scope = "Remote Access Scope"
	remote_access_scope_container = "Remote Access Scope Container"
	report_base = "Report Base"
	report_container = "Report Container"
	report_filter_base = "Report Filter Base"
	report_root = "Report Root"
	report_template = "Report Template"
	report_analytics = "Report:Analytics"
	report_assessor = "Report:Assessor"
	report_certificate_trends_report = "Report:Certificate Trends Report"
	report_certificateduplication = "Report:CertificateDuplication"
	report_certificateinventory = "Report:CertificateInventory"
	report_certificatenetworkvalidation = "Report:CertificateNetworkValidation"
	report_client_certificate_trends_report = "Report:Client Certificate Trends Report"
	report_clients = "Report:Clients"
	report_custom_report = "Report:Custom Report"
	report_demographics_certificate_authority = "Report:Demographics Certificate Authority"
	report_demographics_key_length = "Report:Demographics Key Length"
	report_demographics_signing_algorithm = "Report:Demographics Signing Algorithm"
	report_demographics_validity_period = "Report:Demographics Validity Period"
	report_demographics_wildcard = "Report:Demographics Wildcard"
	report_enhancedcertificateexpiration = "Report:EnhancedCertificateExpiration"
	report_entitlement = "Report:Entitlement"
	report_expiration = "Report:Expiration"
	report_licensing = "Report:Licensing"
	report_ssh_authorized_user = "Report:SSH Authorized User"
	report_ssh_discrepancy = "Report:SSH Discrepancy"
	report_ssh_key_expiration = "Report:SSH Key Expiration"
	report_ssh_server_summary = "Report:SSH Server Summary"
	report_ssh_trust = "Report:SSH Trust"
	report_ssh_user_key = "Report:SSH User Key"
	report_ssh_trends_report = "Report:Ssh Trends Report"
	report_trust = "Report:Trust"
	reporter_service_module = "Reporter Service Module"
	riverbed_steelhead = "Riverbed SteelHead"
	ssh_automatic_rotation_trigger = "SSH Automatic Rotation Trigger"
	ssh_ca_key_pair = "SSH CA Key Pair"
	ssh_ca_key_pair_container = "SSH CA Key Pair Container"
	ssh_ca_template = "SSH CA Template"
	ssh_ca_template_container = "SSH CA Template Container"
	ssh_certificate_adaptable_flow_action = "SSH Certificate Adaptable Flow Action"
	ssh_certificate_create_certificate_action = "SSH Certificate Create Certificate Action"
	ssh_certificate_create_public_key_action = "SSH Certificate Create Public Key Action"
	ssh_certificate_issuance_flow_container = "SSH Certificate Issuance Flow Container"
	ssh_certificate_manager = "SSH Certificate Manager"
	ssh_certificate_start_processing_action = "SSH Certificate Start Processing Action"
	ssh_key = "SSH Key"
	ssh_manager = "SSH Manager"
	ssh_manual_install_key_device = "SSH Manual Install Key Device"
	ssh_server_key = "SSH Server Key"
	ssh_user_key = "SSH User Key"
	ssh_work = "SSH Work"
	sso = "SSO"
	schedule_base = "Schedule Base"
	secret_key = "Secret Key"
	secret_store_root = "Secret Store Root"
	self_signed_ca = "Self Signed CA"
	server_agent_base_device_placement_work = "Server Agent Base Device Placement Work"
	server_agent_cert_device_placement_work = "Server Agent Cert Device Placement Work"
	server_agent_ssh_device_placement_work = "Server Agent SSH Device Placement Work"
	server_certificate_work = "Server Certificate Work"
	service_module = "Service Module"
	ssh_device_base = "Ssh Device Base"
	statistics_container = "Statistics Container"
	statistics_counter = "Statistics Counter"
	statistics_root = "Statistics Root"
	symantec_lhk_ca = "Symantec LHK CA"
	symmetric_key = "Symmetric Key"
	tealeaf_pca = "Tealeaf PCA"
	tealeafsunimp = "TealeafSunimp"
	thawte_spki_ca = "Thawte SPKI CA"
	top = "Top"
	tree_root = "Tree Root"
	trust = "Trust"
	trustwave_ca = "Trustwave CA"
	unicert_ca = "UniCERT CA"
	untrusted_client_group = "Untrusted Client Group"
	upgrades_root = "Upgrades Root"
	user = "User"
	user_preference = "User Preference"
	username_password_credential = "Username Password Credential"
	vam_nshield = "VAM nShield"
	validation_base = "Validation Base"
	validation_manager = "Validation Manager"
	vamcavium = "VamCavium"
	vamsca6000 = "VamSCA6000"
	vamsunimp = "VamSunimp"
	venafi_platform = "Venafi Platform"
	verisign_ca = "VeriSign CA"
	verizon_ca = "Verizon CA"
	workflow = "Workflow"
	workflow_container = "Workflow Container"
	workflow_root = "Workflow Root"
	workflow_ticket = "Workflow Ticket"
	x509_certificate = "X509 Certificate"
	x509_certificate_base = "X509 Certificate Base"
	x509_certificate_validation = "X509 Certificate Validation"
	x509_code_signing_certificate = "X509 Code Signing Certificate"
	x509_device_certificate = "X509 Device Certificate"
	x509_intermediate_root_certificate = "X509 Intermediate Root Certificate"
	x509_root_certificate = "X509 Root Certificate"
	x509_server_certificate = "X509 Server Certificate"
	x509_time_stamping_certificate = "X509 Time Stamping Certificate"
	x509_user_certificate = "X509 User Certificate"
	xolphin_ca = "Xolphin CA"
	zone_base = "Zone Base"
	iplanet = "iPlanet"
	zos_ca = "zOS CA"



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.connection_base import ConnectionBaseAttributes
from pytpp.attributes.ssh_device_base import SshDeviceBaseAttributes
from pytpp.attributes.top import TopAttributes
from pytpp.attributes.validation_base import ValidationBaseAttributes


class DeviceAttributes(ConnectionBaseAttributes, SshDeviceBaseAttributes, TopAttributes, ValidationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Device"
	agentless_discovery_stage = Attribute('Agentless Discovery Stage', min_version='15.2')
	agentless_discovery_status = Attribute('Agentless Discovery Status', min_version='15.2')
	allow_agentless_discovery_and_remediation = Attribute('Allow Agentless Discovery and Remediation', min_version='15.2')
	approver = Attribute('Approver', min_version='17.3')
	bulk_provisioning_dn = Attribute('Bulk Provisioning Dn', min_version='18.3')
	bulk_provisioning_stage = Attribute('Bulk Provisioning Stage', min_version='18.3')
	bulk_provisioning_status = Attribute('Bulk Provisioning Status', min_version='18.3')
	client_id = Attribute('Client ID')
	client_machine_id = Attribute('Client Machine ID', min_version='17.1')
	cloud_instance_id = Attribute('Cloud Instance ID', min_version='17.1')
	cloud_region = Attribute('Cloud Region', min_version='17.1')
	cloud_service = Attribute('Cloud Service', min_version='17.1')
	created_by = Attribute('Created By', min_version='15.1')
	deny_multiple_authentication_failures = Attribute('Deny Multiple Authentication Failures', min_version='19.4')
	disabled_on = Attribute('Disabled On', min_version='17.1')
	discovered_by_dn = Attribute('Discovered By DN', min_version='19.1')
	jump_server_dn = Attribute('Jump Server DN')
	last_attempt_to_get_client_subsystem_record = Attribute('Last Attempt To Get Client Subsystem Record', min_version='17.1')
	last_discovery_date = Attribute('Last Discovery Date', min_version='15.2')
	last_discovery_planned = Attribute('Last Discovery Planned', min_version='16.4')
	last_discovery_platform_version = Attribute('Last Discovery Platform Version', min_version='20.2')
	last_file_operations_platform_version = Attribute('Last File Operations Platform Version', min_version='20.2')
	location = Attribute('Location')
	manual_approval = Attribute('Manual Approval')
	maximum_authorizations_per_keyset = Attribute('Maximum Authorizations Per Keyset', min_version='20.1')
	onboard_discovery_dn = Attribute('Onboard Discovery Dn', min_version='16.1')
	onboard_discovery_stage = Attribute('Onboard Discovery Stage', min_version='16.1')
	onboard_discovery_status = Attribute('Onboard Discovery Status', min_version='16.1')
	onboard_discovery_to_do = Attribute('Onboard Discovery To Do', min_version='16.1')
	placement_job_dn = Attribute('Placement Job Dn', min_version='16.3')
	previous_connection_credential_hash = Attribute('Previous Connection Credential Hash', min_version='19.4')
	privilege_elevation_command = Attribute('Privilege Elevation Command', min_version='17.1')
	protection_key = Attribute('Protection Key')
	required_sync_confirmation = Attribute('Required Sync Confirmation', min_version='19.2')
	ssh_key_encryption = Attribute('SSH Key Encryption', min_version='16.4')
	status = Attribute('Status')
	system_owned_keys = Attribute('System Owned Keys', min_version='15.4')



from pytpp.features.objects import Objects
from pytpp.features.folder import Folder
from pytpp.features.client_groups import (
    Agentless, EstCertificateEnrollment, VenafiAgent
)
from pytpp.features.client_work import (
    AgentConnectivity, AgentUpgrade, CertificateDevicePlacement, CertificateDiscovery,
    CertificateEnrollmentViaESTProtocol, CertificateInstallation, DynamicProvisioning,
    SSHDevicePlacement, SSHDiscovery, SSHKeyUsage, SSHRemediation, UserCertificateCreation
)
from pytpp.features.certificate import Certificate
from pytpp.features.device import Device, JumpServer
from pytpp.features.application import (
    Adaptable as AdapatbleApplication, AmazonAWS, Apache, AzureKeyVault, Basic, BlueCoatSSLVA, CAPI, CitrixNetScaler,
    ConnectDirect, F5AuthenticationBundle, F5LTMAdvanced, GoogleCloudLoadBalancer, IBMDataPower, IBMGSK, ImpervaMX, JKS,
    OracleIPlanet, PaloAltoNetworkFW, PEM, PKCS11, PKCS12, RiverbedSteelHead, TealeafPCA, VAMnShield,
    PKCS11ApplicationGroup, ApacheApplicationGroup
)
from pytpp.features.discovery import NetworkDiscovery
from pytpp.features.credentials import (
    AmazonCredential, CertificateCredential, GenericCredential, PasswordCredential, PrivateKeyCredential,
    UsernamePasswordCredential, GoogleCredential
)
from pytpp.features.certificate_authorities import (MSCA, SelfSignedCA)
from pytpp.features.identity import User, Group
from pytpp.features.permissions import Permissions
from pytpp.features.platforms import Platforms
from pytpp.features.placement_rules import PlacementRules, PlacementRuleCondition
from pytpp.features.workflow import ReasonCode, AdaptableWorkflow, StandardWorkflow, Ticket
from pytpp.features.custom_fields import CustomField


# region Features
class _Application:
    def __init__(self, api):
        self._api = api

        self._adaptable = None
        self._amazon_aws = None
        self._apache = None
        self._azure_key_vault = None
        self._basic = None
        self._blue_coat_sslva = None
        self._capi = None
        self._citrix_net_scaler = None
        self._connect_direct = None
        self._f5_authentication_bundle = None
        self._f5_ltm_advanced = None
        self._google_cloud_load_balancer = None
        self._ibm_datapower = None
        self._ibm_gsk = None
        self._imperva_mx = None
        self._jks = None
        self._oracle_iplanet = None
        self._palo_alto_network_fw = None
        self._pem = None
        self._pkcs11 = None
        self._pkcs12 = None
        self._riverbed_steel_head = None
        self._tealeaf_pca = None
        self._vamnshield = None

    @property
    def adaptable(self) -> AdapatbleApplication:
        self._adaptable = self._adaptable or AdapatbleApplication(self._api)
        return self._adaptable

    @property
    def amazon_aws(self) -> AmazonAWS:
        self._amazon_aws = self._amazon_aws or AmazonAWS(self._api)
        return self._amazon_aws

    @property
    def apache(self) -> Apache:
        self._apache = self._apache or Apache(self._api)
        return self._apache

    @property
    def azure_key_vault(self) -> AzureKeyVault:
        self._azure_key_vault = self._azure_key_vault or AzureKeyVault(self._api)
        return self._azure_key_vault

    @property
    def basic(self) -> Basic:
        self._basic = self._basic or Basic(self._api)
        return self._basic

    @property
    def blue_coat_sslva(self) -> BlueCoatSSLVA:
        self._blue_coat_sslva = self._blue_coat_sslva or BlueCoatSSLVA(self._api)
        return self._blue_coat_sslva

    @property
    def capi(self) -> CAPI:
        self._capi = self._capi or CAPI(self._api)
        return self._capi

    @property
    def citrix_net_scaler(self) -> CitrixNetScaler:
        self._citrix_net_scaler = self._citrix_net_scaler or CitrixNetScaler(self._api)
        return self._citrix_net_scaler

    @property
    def connect_direct(self) -> ConnectDirect:
        self._connect_direct = self._connect_direct or ConnectDirect(self._api)
        return self._connect_direct

    @property
    def f5_authentication_bundle(self) -> F5AuthenticationBundle:
        self._f5_authentication_bundle = self._f5_authentication_bundle or F5AuthenticationBundle(self._api)
        return self._f5_authentication_bundle

    @property
    def f5_ltm_advanced(self) -> F5LTMAdvanced:
        self._f5_ltm_advanced = self._f5_ltm_advanced or F5LTMAdvanced(self._api)
        return self._f5_ltm_advanced

    @property
    def google_cloud_load_balancer(self) -> GoogleCloudLoadBalancer:
        self._google_cloud_load_balancer = self._google_cloud_load_balancer or GoogleCloudLoadBalancer(self._api)
        return self._google_cloud_load_balancer

    @property
    def ibm_datapower(self) -> IBMDataPower:
        self._ibm_datapower = self._ibm_datapower or IBMDataPower(self._api)
        return self._ibm_datapower

    @property
    def ibm_gsk(self) -> IBMGSK:
        self._ibm_gsk = self._ibm_gsk or IBMGSK(self._api)
        return self._ibm_gsk

    @property
    def imperva_mx(self) -> ImpervaMX:
        self._imperva_mx = self._imperva_mx or ImpervaMX(self._api)
        return self._imperva_mx

    @property
    def jks(self) -> JKS:
        self._jks = self._jks or JKS(self._api)
        return self._jks

    @property
    def oracle_iplanet(self) -> OracleIPlanet:
        self._oracle_iplanet = self._oracle_iplanet or OracleIPlanet(self._api)
        return self._oracle_iplanet

    @property
    def palo_alto_network_fw(self) -> PaloAltoNetworkFW:
        self._palo_alto_network_fw = self._palo_alto_network_fw or PaloAltoNetworkFW(self._api)
        return self._palo_alto_network_fw

    @property
    def pem(self) -> PEM:
        self._pem = self._pem or PEM(self._api)
        return self._pem

    @property
    def pkcs11(self) -> PKCS11:
        self._pkcs11 = self._pkcs11 or PKCS11(self._api)
        return self._pkcs11

    @property
    def pkcs12(self) -> PKCS12:
        self._pkcs12 = self._pkcs12 or PKCS12(self._api)
        return self._pkcs12

    @property
    def riverbed_steel_head(self) -> RiverbedSteelHead:
        self._riverbed_steel_head = self._riverbed_steel_head or RiverbedSteelHead(self._api)
        return self._riverbed_steel_head

    @property
    def tealeaf_pca(self) -> TealeafPCA:
        self._tealeaf_pca = self._tealeaf_pca or TealeafPCA(self._api)
        return self._tealeaf_pca

    @property
    def vamnshield(self) -> VAMnShield:
        self._vamnshield = self._vamnshield or VAMnShield(self._api)
        return self._vamnshield


class _ApplicationGroup:
    def __init__(self, api):
        self._api = api

        self._apache = None
        self._pkcs11 = None

    @property
    def apache(self) -> ApacheApplicationGroup:
        self._apache = self._apache or ApacheApplicationGroup(self._api)
        return self._apache

    @property
    def pkcs11(self) -> PKCS11ApplicationGroup:
        self._pkcs11 = self._pkcs11 or PKCS11ApplicationGroup(self._api)
        return self._pkcs11


class _CertificateAuthority:
    def __init__(self, api):
        self._api = api

        self._msca = None
        self._self_signed = None

    @property
    def msca(self) -> MSCA:
        self._msca = self._msca or MSCA(self._api)
        return self._msca

    @property
    def self_signed(self) -> SelfSignedCA:
        self._self_signed = self._self_signed or SelfSignedCA(self._api)
        return self._self_signed


class _ClientGroup:
    def __init__(self, api):
        self._api = api

        self._agentless = None
        self._est_certificate_enrollment = None
        self._venafi_agent = None

    @property
    def agentless(self) -> Agentless:
        self._agentless = self._agentless or Agentless(self._api)
        return self._agentless

    @property
    def est_certificate_enrollment(self) -> EstCertificateEnrollment:
        self._est_certificate_enrollment = self._est_certificate_enrollment or PKCS11ApplicationGroup(self._api)
        return self._est_certificate_enrollment

    @property
    def venafi_agent(self) -> VenafiAgent:
        self._venafi_agent = self._venafi_agent or VenafiAgent(self._api)
        return self._venafi_agent


class _ClientWork:
    def __init__(self, api):
        self._api = api
        self._agent_connectivity = None
        self._agent_upgrade = None
        self._certificate_device_placement = None
        self._certificate_discovery = None
        self._certificate_enrollment_via_est_protocol = None
        self._certificate_installation = None
        self._device_certificate_creation = None
        self._dynamic_provisioning = None
        self._ssh_device_placement = None
        self._ssh_discovery = None
        self._ssh_key_usage = None
        self._ssh_remediation = None
        self._user_certificate_creation = None

    @property
    def agent_connectivity(self) -> AgentConnectivity:
        self._agent_connectivity = self._agent_connectivity or AgentConnectivity(self._api)
        return self._agent_connectivity

    @property
    def agent_upgrade(self) -> AgentUpgrade:
        self._agent_upgrade = self._agent_upgrade or AgentUpgrade(self._api)
        return self._agent_upgrade

    @property
    def certificate_device_placement(self) -> CertificateDevicePlacement:
        self._certificate_device_placement = self._certificate_device_placement or CertificateDevicePlacement(self._api)
        return self._certificate_device_placement

    @property
    def certificate_discovery(self) -> CertificateDiscovery:
        self._certificate_discovery = self._certificate_discovery or CertificateDiscovery(self._api)
        return self._certificate_discovery

    @property
    def certificate_enrollment_via_est_protocol(self) -> CertificateEnrollmentViaESTProtocol:
        self._certificate_enrollment_via_est_protocol = self._certificate_enrollment_via_est_protocol or CertificateEnrollmentViaESTProtocol(
            self._api)
        return self._certificate_enrollment_via_est_protocol

    @property
    def certificate_installation(self) -> CertificateInstallation:
        self._certificate_installation = self._certificate_installation or CertificateInstallation(
            self._api)
        return self._certificate_installation

    @property
    def dynamic_provisioning(self) -> DynamicProvisioning:
        self._dynamic_provisioning = self._dynamic_provisioning or DynamicProvisioning(
            self._api)
        return self._dynamic_provisioning

    @property
    def ssh_device_placement(self) -> SSHDevicePlacement:
        self._ssh_device_placement = self._ssh_device_placement or SSHDevicePlacement(
            self._api)
        return self._ssh_device_placement

    @property
    def ssh_discovery(self) -> SSHDiscovery:
        self._ssh_discovery = self._ssh_discovery or SSHDiscovery(
            self._api)
        return self._ssh_discovery

    @property
    def ssh_key_usage(self) -> SSHKeyUsage:
        self._ssh_key_usage = self._ssh_key_usage or SSHKeyUsage(
            self._api)
        return self._ssh_key_usage

    @property
    def ssh_remediation(self) -> SSHRemediation:
        self._ssh_remediation = self._ssh_remediation or SSHRemediation(
            self._api)
        return self._ssh_remediation

    @property
    def user_certificate_creation(self) -> UserCertificateCreation:
        self._user_certificate_creation = self._user_certificate_creation or UserCertificateCreation(
            self._api)
        return self._user_certificate_creation


class _Credential:
    def __init__(self, api):
        self._api = api

        self._amazon = None
        self._certificate = None
        self._generic = None
        self._google = None
        self._password = None
        self._private_key = None
        self._upcred = None

    @property
    def amazon(self) -> AmazonCredential:
        self._amazon = self._amazon or AmazonCredential(self._api)
        return self._amazon

    @property
    def certificate(self) -> CertificateCredential:
        self._certificate = self._certificate or CertificateCredential(self._api)
        return self._certificate

    @property
    def generic(self) -> GenericCredential:
        self._generic = self._generic or GenericCredential(self._api)
        return self._generic

    @property
    def google(self) -> GoogleCredential:
        self._google = self._google or GoogleCredential(self._api)
        return self._google

    @property
    def password(self) -> PasswordCredential:
        self._password = self._password or PasswordCredential(self._api)
        return self._password

    @property
    def private_key(self) -> PrivateKeyCredential:
        self._private_key = self._private_key or PrivateKeyCredential(self._api)
        return self._private_key

    @property
    def username_password(self) -> UsernamePasswordCredential:
        self._upcred = self._upcred or UsernamePasswordCredential(self._api)
        return self._upcred


class _Discovery:
    def __init__(self, api):
        self._api = api

        self._network = None

    @property
    def network(self) -> NetworkDiscovery:
        self._network = self._network or NetworkDiscovery(api=self._api)
        return self._network


class _Identity:
    def __init__(self, api):
        self._api = api

        self._group = None
        self._user = None

    @property
    def group(self) -> Group:
        self._group = self._group or Group(self._api)
        return self._group

    @property
    def user(self) -> User:
        self._user = self._user or User(self._api)
        return self._user


class _Workflow:
    def __init__(self, api):
        self._api = api

        self._adaptable = None
        self._reason_code = None
        self._standard = None
        self._ticket = None

    @property
    def adaptable(self) -> AdaptableWorkflow:
        self._adaptable = self._adaptable or AdaptableWorkflow(self._api)
        return self._adaptable

    @property
    def reason_code(self) -> ReasonCode:
        self._reason_code = self._reason_code or ReasonCode(self._api)
        return self._reason_code

    @property
    def standard(self) -> StandardWorkflow:
        self._standard = self._standard or StandardWorkflow(self._api)
        return self._standard

    @property
    def ticket(self) -> Ticket:
        self._ticket = self._ticket or Ticket(self._api)
        return self._ticket


class Features:
    def __init__(self, api):
        self._api = api

        self._applications = None
        self._application_group = None
        self._ca = None
        self._certificate = None
        self._client_groups = None
        self._client_work = None
        self._credentials = None
        self._custom_fields = None
        self._device = None
        self._discovery = None
        self._folder = None
        self._identity = None
        self._jump_server = None
        self._objects = None
        self._permissions = None
        self._placement_rule_condition = None
        self._placement_rules = None
        self._platforms = None
        self._workflow = None

    @property
    def application(self) -> _Application:
        self._applications = self._applications or _Application(self._api)
        return self._applications

    @property
    def application_group(self) -> _ApplicationGroup:
        self._application_group = self._application_group or _ApplicationGroup(self._api)
        return self._application_group

    @property
    def certificate(self) -> Certificate:
        self._certificate = self._certificate or Certificate(self._api)
        return self._certificate

    @property
    def certificate_authority(self) -> _CertificateAuthority:
        self._ca = self._ca or _CertificateAuthority(self._api)
        return self._ca

    @property
    def client_groups(self) -> _ClientGroup:
        self._client_groups = self._client_groups or _ClientGroup(self._api)
        return self._client_groups

    @property
    def client_work(self) -> _ClientWork:
        self._client_work = self._client_work or _ClientWork(self._api)
        return self._client_work

    @property
    def credential(self) -> _Credential:
        self._credentials = self._credentials or _Credential(self._api)
        return self._credentials

    @property
    def custom_fields(self) -> CustomField:
        self._custom_fields = self._custom_fields or CustomField(self._api)
        return self._custom_fields

    @property
    def device(self) -> Device:
        self._device = self._device or Device(self._api)
        return self._device

    @property
    def discovery(self) -> _Discovery:
        self._discovery = self._discovery or _Discovery(self._api)
        return self._discovery

    @property
    def folder(self) -> Folder:
        self._folder = self._folder or Folder(self._api)
        return self._folder

    @property
    def identity(self) -> _Identity:
        self._identity = self._identity or _Identity(self._api)
        return self._identity

    @property
    def jump_server(self) -> JumpServer:
        self._jump_server = self._jump_server or JumpServer(self._api)
        return self._jump_server

    @property
    def objects(self) -> Objects:
        self._objects = self._objects or Objects(self._api)
        return self._objects

    @property
    def permissions(self) -> Permissions:
        self._permissions = self._permissions or Permissions(self._api)
        return self._permissions

    @property
    def placement_rule_condition(self) -> PlacementRuleCondition:
        self._placement_rule_condition = self._placement_rule_condition or PlacementRuleCondition()
        return self._placement_rule_condition

    @property
    def placement_rules(self) -> PlacementRules:
        self._placement_rules = self._placement_rules or PlacementRules(self._api)
        return self._placement_rules

    @property
    def platforms(self) -> Platforms:
        self._platforms = self._platforms or Platforms(self._api)
        return self._platforms

    @property
    def workflow(self) -> _Workflow:
        self._workflow = self._workflow or _Workflow(self._api)
        return self._workflow

# endregion



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class FlowAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Flow"
	archive_process = Attribute('Archive Process', min_version='19.2')
	archive_validity = Attribute('Archive Validity', min_version='19.2')
	product_code = Attribute('Product Code', min_version='19.2')



from pytpp.properties.resultcodes import ResultCodes
from pytpp.tools.helpers.date_converter import from_date_string
from pytpp.properties.response_objects.dataclasses import client


class Client:
    @staticmethod
    def Result(code: int):
        return client.Result(
            code=code,
            client_result=ResultCodes.Client.get(code, 'Unknown'),
        )

    @staticmethod
    def Client(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return client.Client(
            client_id=response_object.get('ClientId'),
            client_type=response_object.get('ClientType'),
            fqdn=response_object.get('FQDN'),
            os_name=response_object.get('OsName'),
            username=response_object.get('Username'),
        )

    @staticmethod
    def ClientDetails(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return client.ClientDetails(
            certificate_device=response_object.get('CertificateDevice'),
            client_id=response_object.get('ClientId'),
            client_type=response_object.get('ClientType'),
            client_version=response_object.get('ClientVersion'),
            created_on=from_date_string(response_object.get('CreatedOn')),
            dns_name=response_object.get('DnsName'),
            effective_work=response_object.get('EffectiveWork'),
            fqdn=response_object.get('FQDN'),
            groups=response_object.get('Groups'),
            host_domain=response_object.get('HostDomain'),
            hostname=response_object.get('Hostname'),
            last_seen_on=from_date_string(response_object.get('LastSeenOn')),
            networks=[Client._Network(network) for network in response_object.get('Networks')],
            os_build=response_object.get('OsBuild'),
            os_name=response_object.get('OsName'),
            os_service_pack=response_object.get('OsServicePack'),
            os_version=response_object.get('OsVersion'),
            region=response_object.get('Region'),
            serial_number=response_object.get('SerialNumber'),
            ssh_device=response_object.get('SshDevice'),
            system_architecture=response_object.get('SystemArchitecture'),
            system_chassis=response_object.get('SystemChassis'),
            system_manufacturer=response_object.get('SystemManufacturer'),
            system_model=response_object.get('SystemModel'),
            trust_level=response_object.get('TrustLevel'),
            username=response_object.get('Username'),
            virtual_machine_id=response_object.get('VirtualMachineId'),
        )

    @staticmethod
    def Work(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return client.Work(
            associated_groups=response_object.get('AssociatedGroups'),
            work_dn=response_object.get('WorkDn'),
            work_name=response_object.get('WorkName'),
            work_type=response_object.get('WorkType'),
        )

    @staticmethod
    def _Network(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return client._Network(
            ip_address=response_object.get('IpAddress'),
            mac_address=response_object.get('MacAddress'),
        )



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.driver_base import DriverBaseAttributes


class IdentityDriverAttributes(DriverBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Identity Driver"
	display_name_attributes = Attribute('Display Name Attributes', min_version='15.3')
	mapping_table = Attribute('Mapping Table')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class ClientGroupBaseAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Client Group Base"
	assigned_work = Attribute('Assigned Work')
	client_portal_access_identity = Attribute('Client Portal Access Identity', min_version='15.2')
	fixed_members = Attribute('Fixed Members')
	in_error = Attribute('In Error')
	rank = Attribute('Rank')
	rule = Attribute('Rule')
	status = Attribute('Status')



from pytpp.properties.response_objects.dataclasses import secret_store
from pytpp.properties.resultcodes import ResultCodes


class SecretStore:
    @staticmethod
    def Result(code: int):
        return secret_store.Result(
            code=code,
            secret_store_result=ResultCodes.SecretStore.get(code, 'Unknown'),
        )

    @staticmethod
    def TypedNameValues(response_object: dict):
        if not isinstance(response_object, dict):
            response_object = {}
        return secret_store.TypedNameValues(
            name=response_object.get('Name'),
            type=response_object.get('Type'),
            value=response_object.get('Value'),
        )



from dataclasses import dataclass
from datetime import datetime
from typing import List


@dataclass
class Engine:
    dn: str
    display_name: str
    guid: str
    id: int
    name: str


@dataclass
class Services:
    vplatform: 'Service'
    log_server: 'Service'
    iis: 'Service'


@dataclass
class Service:
    modules: list
    time_since_first_seen: datetime
    time_since_last_seen: datetime
    status: str


@dataclass
class SystemStatus:
    engine_name: str
    services: 'Services'
    version: str


@dataclass
class Task:
    display_name: str
    name: str
    start_time: datetime
    stop_time: datetime
    warning_count: int


@dataclass
class UpgradeInfo:
    id: str
    start_time: datetime
    versions: List[str]


@dataclass
class UpgradeStatus:
    engine: 'Engine'
    status: str
    upgrade_start_time: datetime
    upgrade_stop_time: datetime
    tasks_completed: 'List[Task]'
    tasks_pending: 'List[Task]'
    tasks_running: 'List[Task]'


@dataclass
class UpgradeSummary:
    status: str
    upgrade_start_time: datetime
    upgrade_stop_time: datetime
    completed_tasks: int
    target_version: str
    engines_complete: int
    engines_running: int
    engines_blocked: int
    engines_in_error: int
    engines_pending_install: int



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_base import ApplicationBaseAttributes


class A10AXTrafficManagerAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "A10 AX Traffic Manager"
	bundle_certificate = Attribute('Bundle Certificate')
	certificate_chain_name = Attribute('Certificate Chain Name')
	certificate_name = Attribute('Certificate Name')
	file_validation_disabled = Attribute('File Validation Disabled')
	install_chain_file = Attribute('Install Chain File')
	network_validation_disabled = Attribute('Network Validation Disabled')
	ssh_port = Attribute('SSH Port')
	ssl_cache_size = Attribute('SSL Cache Size')
	ssl_close_notification = Attribute('SSL Close Notification')
	ssl_false_start = Attribute('SSL False Start')
	ssl_pass_phrase = Attribute('SSL Pass Phrase')
	ssl_profile_name = Attribute('SSL Profile Name')
	ssl_profile_type = Attribute('SSL Profile Type')
	tls_version = Attribute('TLS Version')
	use_ssl_template = Attribute('Use SSL Template')



import warnings

class DeprecationMeta(type):
    def __getattribute__(self, item):
        reason = super().__getattribute__('__deprecation_reason__')
        warnings.warn(reason)
        return super().__getattribute__(item)





from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class DriverBaseAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Driver Base"
	driver_arguments = Attribute('Driver Arguments')
	driver_name = Attribute('Driver Name')
	rank = Attribute('Rank')



from pytpp.attributes._helper import IterableMeta, Attribute


class MonitoringBaseAttributes(metaclass=IterableMeta):
	__config_class__ = "Monitoring Base"
	creation_date = Attribute('Creation Date')
	escalation_notice_interval = Attribute('Escalation Notice Interval')
	escalation_notice_start = Attribute('Escalation Notice Start')
	expiration_notice_interval = Attribute('Expiration Notice Interval')
	expiration_notice_start = Attribute('Expiration Notice Start')
	last_notification = Attribute('Last Notification')
	notes = Attribute('Notes')
	notification_disabled = Attribute('Notification Disabled')
	validity_period = Attribute('Validity Period')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class CSPKeyContainerAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "CSP Key Container"
	certificate_vault_id = Attribute('Certificate Vault Id', min_version='19.2')
	options = Attribute('Options', min_version='19.2')
	private_key_vault_id = Attribute('Private Key Vault Id', min_version='19.2')
	private_signing_key_vault_id = Attribute('Private Signing Key Vault Id', min_version='19.2')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.flow_action_base import FlowActionBaseAttributes


class FlowActionApproverBaseAttributes(FlowActionBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Flow Action Approver Base"
	minimum_approvers = Attribute('Minimum Approvers', min_version='19.2')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.http_ca_base import HTTPCABaseAttributes


class DigiCertCAAttributes(HTTPCABaseAttributes, metaclass=IterableMeta):
	__config_class__ = "DigiCert CA"
	api_credentials = Attribute('API Credentials', min_version='17.3')
	api_key = Attribute('API Key')
	account_number = Attribute('Account Number')
	account_organization = Attribute('Account Organization', min_version='17.3')
	allow_reissue = Attribute('Allow Reissue', min_version='17.3')
	certificate_transparency = Attribute('Certificate Transparency', min_version='18.4')
	division = Attribute('Division', min_version='20.3')
	ev_allowed = Attribute('EV Allowed')
	ev_enabled = Attribute('EV Enabled')
	manual_approval = Attribute('Manual Approval')
	organizational_unit = Attribute('Organizational Unit', min_version='17.3')
	profile_id = Attribute('Profile ID', min_version='15.1')
	renewal_window = Attribute('Renewal Window', min_version='17.3')
	san_enabled = Attribute('SAN Enabled')
	uc_allowed = Attribute('UC Allowed')
	wildcard_allowed = Attribute('Wildcard Allowed')



from typing import List
from pytpp.api.api_base import API, APIResponse, api_response_property
from pytpp.properties.response_objects.certificate import Certificate
from pytpp.tools.helpers.date_converter import from_date_string


class _Certificates(API):
    def __init__(self, api_obj):
        super().__init__(api_obj=api_obj, url='/Certificates')
        self.Associate = self._Associate(api_obj=api_obj)
        self.CheckPolicy = self._CheckPolicy(api_obj=api_obj)
        self.Dissociate = self._Dissociate(api_obj=api_obj)
        self.Import = self._Import(api_obj=api_obj)
        self.Push = self._Push(api_obj=api_obj)
        self.Renew = self._Renew(api_obj=api_obj)
        self.Request = self._Request(api_obj=api_obj)
        self.Reset = self._Reset(api_obj=api_obj)
        self.Retrieve = self._Retrieve(api_obj=api_obj)
        self.Retry = self._Retry(api_obj=api_obj)
        self.Revoke = self._Revoke(api_obj=api_obj)
        self.Validate = self._Validate(api_obj=api_obj)

    def Guid(self, guid):
        return self._Guid(guid=guid, api_obj=self._api_obj)

    def get(self, limit: int = None, offset: int = None, optional_fields: list = None, filters: dict = None):
        params = {
            'Limit': limit,
            'Offset': offset,
            'OptionalFields': optional_fields
        }
        params.update(filters or {})

        class _Response(APIResponse):
            def __init__(self, response):
                super().__init__(response=response)

            @property
            @api_response_property()
            def links(self):
                links = self._from_json(key='_links')
                return [Certificate.Link(lnk) for lnk in links]

            @property
            @api_response_property()
            def x_record_count(self) -> int:
                xrc = self.api_response.headers.get('X-Record-Count')
                return xrc

            @property
            @api_response_property()
            def certificates(self):
                certs = self._from_json(key='Certificates')
                return [Certificate.Certificate(cert) for cert in certs]

            @property
            @api_response_property()
            def data_range(self) -> str:
                return self._from_json(key='DataRange')

            @property
            @api_response_property()
            def total_count(self) -> int:
                return self._from_json(key='TotalCount')

        return _Response(response=self._get(params=params))

    class _Associate(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Certificates/Associate')

        def post(self, application_dn: str, certificate_dn: str, push_to_new: bool):
            body = {
                'CertificateDN': certificate_dn,
                'ApplicationDN': application_dn,
                'PushToNew': push_to_new
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def success(self) -> bool:
                    return self._from_json(key='Success')

            return _Response(response=self._post(data=body))

    class _CheckPolicy(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Certificates/CheckPolicy')

        def post(self, policy_dn: str, pkcs10: str = None):
            body = {
                'PolicyDN': policy_dn,
                'PKSC10': pkcs10
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def csr(self):
                    return Certificate.CSR(self._from_json(key='CSR', error_key='Error'))

                @property
                @api_response_property()
                def policy(self):
                    return Certificate.Policy(self._from_json(key='Policy'))

            return _Response(response=self._post(data=body))

    class _Dissociate(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Certificates/Dissociate')

        def post(self, certificate_dn: str, application_dn: list, delete_orphans: bool = False):
            body = {
                'CertificateDN': certificate_dn,
                'ApplicationDN': application_dn,
                'DeleteOrphans': delete_orphans
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def success(self) -> bool:
                    return self._from_json(key='Success')

            return _Response(response=self._post(data=body))

    class _Guid(API):
        def __init__(self, guid: str, api_obj):
            self._cert_guid = guid
            super().__init__(api_obj=api_obj, url='/Certificates/{guid}'.format(guid=self._cert_guid))
            self.PreviousVersions = self._PreviousVersions(guid=self._cert_guid, api_obj=api_obj)
            self.ValidationResults = self._ValidationResults(guid=self._cert_guid, api_obj=api_obj)

        def delete(self):
            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def success(self) -> bool:
                    return self._from_json(key='Success')

            return _Response(response=self._delete())

        def get(self):
            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def approver(self) -> List[str]:
                    return self._from_json(key='Approver')

                @property
                @api_response_property()
                def certificate_details(self):
                    return Certificate.CertificateDetails(self._from_json(key='CertificateDetails'))

                @property
                @api_response_property()
                def contact(self) -> List[str]:
                    return self._from_json(key='Contact')

                @property
                @api_response_property()
                def created_on(self):
                    return from_date_string(self._from_json(key='CreatedOn'))

                @property
                @api_response_property()
                def custom_fields(self) -> List[dict]:
                    return self._from_json(key='CustomFields')

                @property
                @api_response_property()
                def dn(self) -> str:
                    return self._from_json(key='DN')

                @property
                @api_response_property()
                def guid(self) -> str:
                    return self._from_json(key='Guid')

                @property
                @api_response_property()
                def name(self) -> str:
                    return self._from_json(key='Name')

                @property
                @api_response_property()
                def parent_dn(self) -> str:
                    return self._from_json(key='ParentDN')

                @property
                @api_response_property()
                def processing_details(self):
                    return Certificate.ProcessingDetails(self._from_json(key='ProcessingDetails'))

                @property
                @api_response_property()
                def renewal_details(self):
                    return Certificate.RenewalDetails(self._from_json(key='RenewalDetails'))

                @property
                @api_response_property()
                def schema_class(self) -> str:
                    return self._from_json(key='SchemaClass')

                @property
                @api_response_property()
                def success(self) -> str:
                    return self._from_json(key='Success')

                @property
                @api_response_property()
                def validation_details(self):
                    return Certificate.ValidationDetails(self._from_json(key='ValidationDetails'))

            return _Response(response=self._get())

        def put(self, attribute_data: List[dict]):
            body = {
                "AttributeData": attribute_data
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def success(self) -> str:
                    return self._from_json(key='Success')

            return _Response(response=self._put(data=body))

        class _PreviousVersions(API):
            def __init__(self, guid: str, api_obj):
                self._cert_guid = guid
                super().__init__(
                    api_obj=api_obj,
                    url='/Certificates/{guid}/PreviousVersions'.format(guid=self._cert_guid)
                )

            def get(self, exclude_expired: bool = False, exclude_revoked: bool = False):
                params = {
                    'ExcludeExpired': exclude_expired,
                    'ExcludeRevoked': exclude_revoked
                }
                
                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response)

                    @property
                    @api_response_property()
                    def success(self) -> bool:
                        return self._from_json(key='Success')

                    @property
                    @api_response_property()
                    def previous_versions(self):
                        return [Certificate.PreviousVersions(version) for version in self._from_json(key='PreviousVersions')]

                return _Response(response=self._get(params=params))

        class _ValidationResults(API):
            def __init__(self, guid: str, api_obj):
                self._cert_guid = guid
                super().__init__(
                    api_obj=api_obj,
                    url='/Certificates/{guid}/ValidationResults'.format(guid=self._cert_guid)
                )

            def get(self):
                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response)

                    @property
                    @api_response_property(return_on_204=list)
                    def file(self):
                        return [Certificate.File(f) for f in self._from_json(key='File')]

                    @property
                    @api_response_property(return_on_204=list)
                    def ssl_tls(self):
                        return [Certificate.SslTls(s) for s in self._from_json(key='SslTls')]

                return _Response(response=self._get())

    class _Import(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Certificates/Import')

        def post(self, certificate_data: str, policy_dn: str, ca_specific_attributes: list = None, object_name: str = None,
                 password: str = None, private_key_data: str = None, reconcile: bool = False):
            body = {
                'CertificateData': certificate_data,
                'PolicyDN': policy_dn,
                'CASpecificAttributes': ca_specific_attributes,
                'ObjectName': object_name,
                'Password': password,
                'PrivateKeyData': private_key_data,
                'Reconcile': reconcile
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def certificate_dn(self) -> str:
                    return self._from_json(key='CertificateDN')

                @property
                @api_response_property()
                def certificate_vault_id(self) -> int:
                    return self._from_json(key='CertificateVaultID')

                @property
                @api_response_property()
                def guid(self) -> str:
                    return self._from_json(key='Guid')

                @property
                @api_response_property()
                def private_key_vault_id(self) -> int:
                    return self._from_json(key='PrivateKeyVaultID')

            return _Response(response=self._post(data=body))

    class _Push(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='Certificates/Push')

        def post(self, certificate_dn: str, application_dn: List[str] = None, push_to_all: bool = False):
            body = {
                'ApplicationDN': application_dn,
                'CertificateDN': certificate_dn,
                'PushToAll': push_to_all
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def success(self) -> bool:
                    return self._from_json(key='Success')

            return _Response(response=self._post(data=body))

    class _Renew(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Certificates/Renew')

        def post(self, certificate_dn: str, pkcs10: str = None, reenable: bool = False):
            body = {
                'CertificateDN': certificate_dn,
                'PKCS10': pkcs10,
                'Reenable': reenable
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def success(self) -> bool:
                    return self._from_json(key='Success')

            return _Response(response=self._post(data=body))

    class _Request(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Certificates/Request')

        def post(self, policy_dn: str, approvers: List[dict] = None, cadn: str = None,
                 ca_specific_attributes: List[dict] = None, certificate_type: str = None, city: str = None,
                 contacts: List[dict] = None, country: str=None, custom_fields: List[dict] = None, created_by: str = None,
                 devices: List[dict] = None, disable_automatic_renewal: bool = False, elliptic_curve: str = None,
                 key_algorithm: str = None, key_bit_size: int = None, management_type: str = None, object_name: str = None,
                 organization: str = None, organizational_unit: str = None, origin: str = None, pkcs10: str = None,
                 reenable: bool = False, set_work_todo: bool = True, state: str = None, subject: str = None,
                 subject_alt_names: List[dict] = None):
            body = {
                'Approvers'              : approvers,
                'CADN'                   : cadn,
                'CASpecificAttributes'   : ca_specific_attributes,
                'CertificateType'        : certificate_type,
                'City'                   : city,
                'Contacts'               : contacts,
                'Country'                : country,
                'CustomFields'           : custom_fields,
                'CreatedBy'              : created_by,
                'Devices'                : devices,
                'DisableAutomaticRenewal': disable_automatic_renewal,
                'EllipticCurve'          : elliptic_curve,
                'KeyAlgorithm'           : key_algorithm,
                'KeyBitSize'             : key_bit_size,
                'ManagementType'         : management_type,
                'ObjectName'             : object_name,
                'Origin'                 : origin,
                'Organization'           : organization,
                'OrganizationalUnit'     : organizational_unit,
                'PKCS10'                 : pkcs10,
                'PolicyDN'               : policy_dn,
                'Reenable'               : reenable,
                'SetWorkToDo'            : set_work_todo,
                'State'                  : state,
                'Subject'                : subject,
                'SubjectAltNames'        : subject_alt_names,
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def certificate_dn(self) -> str:
                    return self._from_json(key='CertificateDN')

                @property
                @api_response_property()
                def guid(self) -> str:
                    return self._from_json(key='Guid')

            return _Response(response=self._post(data=body))

    class _Reset(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Certificates/Reset')

        def post(self, certificate_dn: str, restart: bool = False):
            body = {
                'CertificateDN': certificate_dn,
                'Restart': restart
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def private_key_mismatch_reset_completed(self) -> bool:
                    return self._from_json(key='PrivateKeyMismatchResetCompleted')

                @property
                @api_response_property()
                def processing_reset_completed(self) -> bool:
                    return self._from_json(key='ProcessingResetCompleted')

                @property
                @api_response_property()
                def restart_completed(self) -> bool:
                    return self._from_json(key='RestartCompleted')

                @property
                @api_response_property()
                def revocation_reset_completed(self) -> bool:
                    return self._from_json(key='RevocationResetCompleted')

            return _Response(response=self._post(data=body))

    class _Retrieve(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Certificates/Retrieve')

        def get(self, certificate_dn: str, format: str, friendly_name: str, include_chain: bool = False,
                include_private_key: bool = False, keystore_password: str = None, password: str = None,
                root_first_order: bool = False):
            params = {
                'CertificateDN': certificate_dn,
                'Format': format,
                'FriendlyName': friendly_name,
                'IncludeChain': include_chain,
                'IncludePrivateKey': include_private_key,
                'KeystorePassword': keystore_password,
                'Password': password,
                'RootFirstOrder': root_first_order
            }

            return APIResponse(response=self._get(params=params))

        def post(self, certificate_dn: str, format: str, friendly_name: str, include_chain: bool = False,
                 include_private_key: bool = False, keystore_password: str = None, password: str = None,
                 root_first_order: bool = False):
            body = {
                'CertificateDN': certificate_dn,
                'Format': format,
                'FriendlyName': friendly_name,
                'IncludeChain': include_chain,
                'IncludePrivateKey': include_private_key,
                'KeystorePassword': keystore_password,
                'Password': password,
                'RootFirstOrder': root_first_order
            }
            
            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def certificate_data(self) -> str:
                    return self._from_json(key='CertificateData')

                @property
                @api_response_property()
                def filename(self) -> str:
                    return self._from_json(key='Filename')

                @property
                @api_response_property()
                def format(self) -> str:
                    return self._from_json(key='Format')

            return _Response(response=self._post(data=body))

        def VaultId(self, vault_id: int):
            return self._VaultId(vault_id=vault_id, api_obj=self._api_obj)

        class _VaultId(API):
            def __init__(self, vault_id: int, api_obj):
                super().__init__(api_obj=api_obj, url='/Certificates/Retrieve/{vault_id}'.format(vault_id=vault_id))
                self._vault_id = vault_id

            def get(self, format: str, friendly_name: str, include_chain: bool = False,
                    include_private_key: bool = False, keystore_password: str = None, password: str = None,
                    root_first_order: bool = False):
                params = {
                    'Format': format,
                    'FriendlyName': friendly_name,
                    'IncludeChain': include_chain,
                    'IncludePrivateKey': include_private_key,
                    'KeystorePassword': keystore_password,
                    'Password': password,
                    'RootFirstOrder': root_first_order
                }

                return APIResponse(response=self._get(params=params))

            def post(self, format: str, friendly_name: str, include_chain: bool = False,
                     include_private_key: bool = False, keystore_password: str = None, password: str = None,
                     root_first_order: bool = False):
                body = {
                    'Format': format,
                    'FriendlyName': friendly_name,
                    'IncludeChain': include_chain,
                    'IncludePrivateKey': include_private_key,
                    'KeystorePassword': keystore_password,
                    'Password': password,
                    'RootFirstOrder': root_first_order
                }

                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response)

                    @property
                    @api_response_property()
                    def certificate_data(self) -> str:
                        return self._from_json(key='CertificateData')

                    @property
                    @api_response_property()
                    def filename(self) -> str:
                        return self._from_json(key='Filename')

                    @property
                    @api_response_property()
                    def format(self) -> str:
                        return self._from_json(key='Format')

                return _Response(response=self._post(data=body))

    class _Retry(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Certificates/Retry')

        def post(self, certificate_dn: str):
            body = {
                'CertificateDN': certificate_dn
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def success(self) -> bool:
                    return self._from_json(key='Success')

            return _Response(response=self._post(data=body))

    class _Revoke(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Certificates/Revoke')

        def post(self, certificate_dn: str = None, thumbprint: str = None, reason: str = None, comments: str = None,
                 disable: bool = None):
            body = {
                'CertificateDN': certificate_dn,
                'Thumbprint': thumbprint,
                'Reason': reason,
                'Comments': comments,
                'Disable': disable
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def error(self) -> bool:
                    return self._from_json(key='Error')

                @property
                @api_response_property()
                def requested(self) -> bool:
                    return self._from_json(key='Requested')

                @property
                @api_response_property()
                def success(self) -> bool:
                    return self._from_json(key='Success')

            return _Response(response=self._post(data=body))

    class _Validate(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='/Certificates/Validate')

        def post(self, certificate_dns: List[str] = None, certificate_guids: List[str] = None):
            body = {
                'CertificateDNs': certificate_dns,
                'CertificateGUIDs': certificate_guids
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def error(self) -> str:
                    return self._from_json(key='Error')

                @property
                @api_response_property()
                def success(self) -> bool:
                    return self._from_json(key='Success')

                @property
                @api_response_property()
                def validated_certificate_dns(self) -> List[str]:
                    return self._from_json(key='ValidatedCertificateDNs')

                @property
                @api_response_property()
                def validated_certificate_guids(self) -> List[str]:
                    return self._from_json(key='ValidatedCertificateGUIDs')

                @property
                @api_response_property()
                def warnings(self) -> List[str]:
                    return self._from_json(key='Warnings')

            return _Response(response=self._post(data=body))



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class LogHeartbeatAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Log Heartbeat"
	event = Attribute('Event')
	rule = Attribute('Rule')
	timeout = Attribute('Timeout')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.http_ca_base import HTTPCABaseAttributes


class GeotrustCAAttributes(HTTPCABaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Geotrust CA"
	account_type = Attribute('Account Type')
	address = Attribute('Address')
	billing_contact_first_name = Attribute('Billing Contact First Name')
	billing_contact_internet_email_address = Attribute('Billing Contact Internet EMail Address')
	billing_contact_last_name = Attribute('Billing Contact Last Name')
	billing_contact_telephone_number = Attribute('Billing Contact Telephone Number')
	city = Attribute('City')
	country = Attribute('Country')
	interval = Attribute('Interval')
	organization = Attribute('Organization')
	partner_code = Attribute('Partner Code')
	postal_code = Attribute('Postal Code')
	state = Attribute('State')
	technical_contact_address = Attribute('Technical Contact Address')
	technical_contact_city = Attribute('Technical Contact City')
	technical_contact_country = Attribute('Technical Contact Country')
	technical_contact_first_name = Attribute('Technical Contact First Name')
	technical_contact_internet_email_address = Attribute('Technical Contact Internet EMail Address')
	technical_contact_last_name = Attribute('Technical Contact Last Name')
	technical_contact_organization = Attribute('Technical Contact Organization')
	technical_contact_postal_code = Attribute('Technical Contact Postal Code')
	technical_contact_state = Attribute('Technical Contact State')
	technical_contact_telephone_number = Attribute('Technical Contact Telephone Number')
	technical_contact_title = Attribute('Technical Contact Title')
	telephone_number = Attribute('Telephone Number')
	test_account = Attribute('Test Account')
	web_service_url = Attribute('Web Service URL')



from pytpp.api.api_base import API, APIResponse, api_response_property
from pytpp.properties.response_objects.flow import Flow
from pytpp.properties.response_objects.codesign import CodeSign
from typing import List


class _Flow:
    def __init__(self, api_obj):
        self.Actions = self._Actions(api_obj=api_obj)
        self.Tickets = self._Tickets(api_obj=api_obj)

    class _Actions:
        def __init__(self, api_obj):
            self.CodeSign = self._CodeSign(api_obj=api_obj)

        class _CodeSign:
            def __init__(self, api_obj):
                self.PreQualify = self._PreQualify(api_obj=api_obj)

            class _PreQualify:
                def __init__(self, api_obj):
                    self.Create = self._Create(api_obj=api_obj)

                class _Create(API):
                    def __init__(self, api_obj):
                        super().__init__(api_obj=api_obj, url='Flow/Actions/CodeSign/PreQualify/Create')

                    def post(self, comment: str, data: str, dn: str, user: str, hours: int = None,
                             single_use: bool = None):
                        body = {
                            'Comment': comment,
                            'Data': data,
                            'Dn': dn,
                            'Hours': hours,
                            'SingleUse': single_use,
                            'User': user
                        }

                        class _Response(APIResponse):
                            def __init__(self, response):
                                super().__init__(response=response)

                            @property
                            @api_response_property()
                            def error(self) -> str:
                                return self._from_json(key='Error')

                            @property
                            @api_response_property()
                            def result(self):
                                return CodeSign.ResultCode(self._from_json(key='Result'))

                            @property
                            @api_response_property()
                            def success(self) -> bool:
                                return self._from_json(key='Success')

                        return _Response(response=self._post(data=body))

    class _Tickets:
        def __init__(self, api_obj):
            self.Approve = self._Approve(api_obj=api_obj)
            self.Count = self._Count(api_obj=api_obj)
            self.CountApproved = self._CountApproved(api_obj=api_obj)
            self.Enumerate = self._Enumerate(api_obj=api_obj)
            self.EnumerateApproved = self._EnumerateApproved(api_obj=api_obj)
            self.Load = self._Load(api_obj=api_obj)
            self.Reject = self._Reject(api_obj=api_obj)
            self.Update = self._Update(api_obj=api_obj)

        class _Approve(API):
            def __init__(self, api_obj):
                super().__init__(api_obj=api_obj, url='/Flow/Tickets/Approve')

            def post(self, ticket_id: int = None, ticket_ids: List[int] = None, expires: str = None,
                     comment: str = None, not_before: str = None, use_count: int = None):
                body = {
                    'Expires'  : expires,
                    'Comment'  : comment,
                    'notBefore': not_before,
                    'TicketId' : ticket_id,
                    'TicketIds': ticket_ids,
                    'useCount' : use_count
                }

                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response)

                    @property
                    @api_response_property()
                    def result(self):
                        return Flow.Result(self._from_json(key='Result'))

                    @property
                    @api_response_property()
                    def message(self) -> str:
                        return self._from_json(key='Message')

                return _Response(response=self._post(data=body))

        class _Count(API):
            def __init__(self, api_obj):
                super().__init__(api_obj=api_obj, url='/Flow/Tickets/Count')

            def post(self):
                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response)

                    @property
                    @api_response_property()
                    def count(self) -> int:
                        return self._from_json(key='Count')

                    @property
                    @api_response_property()
                    def message(self) -> str:
                        return self._from_json(key='Message')

                    @property
                    @api_response_property()
                    def result(self):
                        return Flow.Result(self._from_json(key='Result'))

                return _Response(response=self._post(data={}))

        class _CountApproved(API):
            def __init__(self, api_obj):
                super().__init__(api_obj=api_obj, url='/Flow/Tickets/CountApproved')

            def post(self):
                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response)

                    @property
                    @api_response_property()
                    def count(self) -> int:
                        return self._from_json(key='Count')

                    @property
                    @api_response_property()
                    def message(self) -> str:
                        return self._from_json(key='Message')

                    @property
                    @api_response_property()
                    def result(self):
                        return Flow.Result(self._from_json(key='Result'))

                return _Response(response=self._post(data={}))

        class _Enumerate(API):
            def __init__(self, api_obj):
                super().__init__(api_obj=api_obj, url='/Flow/Tickets/Enumerate')

            def post(self, product_code: int = None, ticket_page_size: int = None, ticket_page_number: int = None):
                body = {
                    'ProductCode'     : product_code,
                    'TicketPageSize'  : ticket_page_size,
                    'TicketPageNumber': ticket_page_number
                }

                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response)

                    @property
                    @api_response_property()
                    def result(self):
                        return Flow.Result(self._from_json(key='Result'))

                    @property
                    @api_response_property()
                    def message(self) -> str:
                        return self._from_json(key='Message')

                    @property
                    @api_response_property()
                    def tickets(self):
                        return [Flow.Ticket(ticket) for ticket in self._from_json(key='Tickets')]

                return _Response(response=self._post(data=body))

        class _EnumerateApproved(API):
            def __init__(self, api_obj):
                super().__init__(api_obj=api_obj, url='/Flow/Tickets/EnumerateApproved')

            def post(self, product_code: int = None, ticket_page_size: int = None, ticket_page_number: int = None):
                body = {
                    'ProductCode'     : product_code,
                    'TicketPageSize'  : ticket_page_size,
                    'TicketPageNumber': ticket_page_number
                }

                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response)

                    @property
                    @api_response_property()
                    def result(self):
                        return Flow.Result(self._from_json(key='Result'))

                    @property
                    @api_response_property()
                    def message(self) -> str:
                        return self._from_json(key='Message')

                    @property
                    @api_response_property()
                    def tickets(self):
                        return [Flow.Ticket(ticket) for ticket in self._from_json(key='Tickets')]

                return _Response(response=self._post(data=body))

        class _Load(API):
            def __init__(self, api_obj):
                super().__init__(api_obj=api_obj, url='/Flow/Tickets/Load')

            def post(self, ticket_id: int = None, ticket_ids: List[int] = None):
                body = {
                    'TicketId' : ticket_id,
                    'TicketIds': ticket_ids
                }

                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response)

                    @property
                    @api_response_property()
                    def result(self):
                        return Flow.Result(self._from_json(key='Result'))

                    @property
                    @api_response_property()
                    def message(self) -> str:
                        return self._from_json(key='Message')

                    @property
                    @api_response_property()
                    def tickets(self):
                        return [Flow.Ticket(ticket) for ticket in self._from_json(key='Tickets')]

                return _Response(response=self._post(data=body))

        class _Reject(API):
            def __init__(self, api_obj):
                super().__init__(api_obj=api_obj, url='/Flow/Tickets/Reject')

            def post(self, comment: str, rejection_level: int, ticket_id: int = None,
                     ticket_ids: List[int] = None):
                body = {
                    'Comment'       : comment,
                    'RejectionLevel': rejection_level,
                    'TicketId'      : ticket_id,
                    'TicketIds'     : ticket_ids
                }

                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response)

                    @property
                    @api_response_property()
                    def invalid_ticket_ids(self) -> List[int]:
                        return self._from_json(key='InvalidTicketIds')

                    @property
                    @api_response_property()
                    def message(self) -> str:
                        return self._from_json(key='Message')

                    @property
                    @api_response_property()
                    def result(self):
                        return Flow.Result(self._from_json(key='Result'))

                return _Response(response=self._post(data=body))

        class _Update(API):
            def __init__(self, api_obj):
                super().__init__(api_obj=api_obj, url='/Flow/Tickets/Update')

            def post(self, ticket_id: int = None, ticket_ids: List[int] = None, expires: str = None,
                     comment: str = None, not_before: str = None, use_count: int = None):
                body = {
                    'Expires'  : expires,
                    'Comment'  : comment,
                    'notBefore': not_before,
                    'TicketId' : ticket_id,
                    'TicketIds': ticket_ids,
                    'useCount' : use_count
                }

                class _Response(APIResponse):
                    def __init__(self, response):
                        super().__init__(response=response)

                    @property
                    @api_response_property()
                    def result(self):
                        return Flow.Result(self._from_json(key='Result'))

                    @property
                    @api_response_property()
                    def message(self) -> str:
                        return self._from_json(key='Message')

                return _Response(response=self._post(data=body))



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.connection_base import ConnectionBaseAttributes
from pytpp.attributes.driver_base import DriverBaseAttributes
from pytpp.attributes.schedule_base import ScheduleBaseAttributes


class BulkApplicationBaseAttributes(ConnectionBaseAttributes, DriverBaseAttributes, ScheduleBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Bulk Application Base"
	batch_size = Attribute('Batch Size', min_version='20.1')
	certificate_thumbprint = Attribute('Certificate Thumbprint', min_version='18.3')
	device = Attribute('Device', min_version='18.3')
	exclude_expired_certificates = Attribute('Exclude Expired Certificates', min_version='18.3')
	exclude_historical_certificates = Attribute('Exclude Historical Certificates', min_version='18.3')
	exclude_revoked_certificates = Attribute('Exclude Revoked Certificates', min_version='18.3')
	grouping_id = Attribute('Grouping Id', min_version='18.3')
	in_error = Attribute('In Error', min_version='18.3')
	in_progress = Attribute('In Progress', min_version='18.3')
	last_run = Attribute('Last Run', min_version='18.3')
	last_update = Attribute('Last Update', min_version='18.3')
	light_run = Attribute('Light Run', min_version='18.3')
	light_run_new_certificates_threshold = Attribute('Light Run New Certificates Threshold', min_version='18.3')
	maximum_days_expired = Attribute('Maximum Days Expired', min_version='18.3')
	policydn = Attribute('PolicyDN', min_version='18.3')
	status = Attribute('Status', min_version='18.3')
	stop_requested = Attribute('Stop Requested', min_version='18.3')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.http_ca_base import HTTPCABaseAttributes
from pytpp.attributes.proxy import ProxyAttributes


class VerizonCAAttributes(HTTPCABaseAttributes, ProxyAttributes, metaclass=IterableMeta):
	__config_class__ = "Verizon CA"
	admin_email = Attribute('Admin Email')
	admin_firstname = Attribute('Admin Firstname')
	admin_surname = Attribute('Admin Surname')
	admin_telnumber = Attribute('Admin Telnumber')
	certificate_block = Attribute('Certificate Block')



from typing import List, Dict
from pytpp.features.bases.feature_base import FeatureBase, feature


# region Platform Components
class _PlatformComponentBase(FeatureBase):
    def __init__(self, api, module: str):
        super().__init__(api=api)
        self._engines_dn = r'\VED\Engines'
        self._module = module

    def _module_dn(self, engine_name: str):
        return f'{self._engines_dn}\\{engine_name}\\{self._module}'

    def _engine_has_module_enabled(self, engine_name: str):
        engines = self._api.websdk.SystemStatus.get().engines
        for engine in engines:
            if engine.engine_name.lower() == engine_name.lower():
                if self._module in engine.services.vplatform.modules:
                    return True
        return False

    def update_engines(self, attributes: Dict[str, List[str]], engine_names: List[str] = None):
        """
        Updates a Platform's module attributes. Each engine in ``engine_names`` will be updated
        to have the given ``attributes``.

        Args:
            attributes: Dictionary of attributes and attribute values to update.
            engine_names: List of engine names.
        """
        engine_names = engine_names or [engine.engine_name for engine in self._api.websdk.ProcessingEngines.get().engines]
        for engine_name in engine_names:
            response = self._api.websdk.Config.Write.post(
                object_dn=self._module_dn(engine_name=engine_name),
                attribute_data=self._name_value_list(attributes, keep_list_values=True)
            )
            response.assert_valid_response()


@feature('Auto Layout Manager')
class AutoLayoutManager(_PlatformComponentBase):
    def __init__(self, api):
        super().__init__(api=api, module='Auto Layout Manager')


@feature('Bulk Provisioning Manager')
class BulkProvisioningManager(_PlatformComponentBase):
    def __init__(self, api):
        super().__init__(api=api, module='Bulk Provisioning Manager')


@feature('CA Import Manager')
class CAImportManager(_PlatformComponentBase):
    def __init__(self, api):
        super().__init__(api=api, module='CA Import Manager')


@feature('Certificate Manager')
class CertificateManager(_PlatformComponentBase):
    def __init__(self, api):
        super().__init__(api=api, module='Certificate Manager')


@feature('Certificate Pre-Enrollment')
class CertificatePreEnrollment(_PlatformComponentBase):
    def __init__(self, api):
        super().__init__(api=api, module='Certificate Pre-Enrollment')


@feature('Certificate Revocation')
class CertificateRevocation(_PlatformComponentBase):
    def __init__(self, api):
        super().__init__(api=api, module='Certificate Revocation')


@feature('Cloud Instance Monitor')
class CloudInstanceMonitor(_PlatformComponentBase):
    def __init__(self, api):
        super().__init__(api=api, module='Cloud Instance Monitor')


@feature('Discovery Manager')
class DiscoveryManager(_PlatformComponentBase):
    def __init__(self, api):
        super().__init__(api=api, module='Discovery')


@feature('Monitor')
class Monitor(_PlatformComponentBase):
    def __init__(self, api):
        super().__init__(api=api, module='Monitor')


@feature('Onboard Discovery Manager')
class OnboardDiscoveryManager(_PlatformComponentBase):
    def __init__(self, api):
        super().__init__(api=api, module='Onboard Discovery Manager')


@feature('Reporting')
class Reporting(_PlatformComponentBase):
    def __init__(self, api):
        super().__init__(api=api, module='Reporting')


@feature('SSH Manager')
class SSHManager(_PlatformComponentBase):
    def __init__(self, api):
        super().__init__(api=api, module='SSH Manager')


@feature('TrustNet Manager')
class TrustNetManager(_PlatformComponentBase):
    def __init__(self, api):
        super().__init__(api=api, module='TrustNet Manager')


@feature('Validation Manager')
class ValidationManager(_PlatformComponentBase):
    def __init__(self, api):
        super().__init__(api=api, module='Validation Manager')
# endregion Platform Components


# region Platform Root
@feature('Platforms')
class Platforms(FeatureBase):
    def __init__(self, api):
        super().__init__(api=api)
        self._engines_dn = r'\VED\Engines'

        self._auto_layout_manager = None
        self._bulk_provisioning_manager = None
        self._ca_import_manager = None
        self._certificate_manager = None
        self._certificate_pre_enrollment = None
        self._certificate_revocation = None
        self._cloud_instance_monitor = None
        self._discovery_manager = None
        self._monitor = None
        self._onboard_discovery_manager = None
        self._reporting = None
        self._ssh_manager = None
        self._trustnet_manager = None
        self._validation_manager = None

    # region Component Properties
    @property
    def auto_layout_manager(self) -> AutoLayoutManager:
        self._auto_layout_manager = self._auto_layout_manager or AutoLayoutManager(self._api)
        return self._auto_layout_manager

    @property
    def bulk_provisioning_manager(self) -> BulkProvisioningManager:
        self._bulk_provisioning_manager = self._bulk_provisioning_manager or BulkProvisioningManager(self._api)
        return self._bulk_provisioning_manager

    @property
    def ca_import_manager(self) -> CAImportManager:
        self._ca_import_manager = self._ca_import_manager or CAImportManager(self._api)
        return self._ca_import_manager

    @property
    def certificate_manager(self) -> CertificateManager:
        self._certificate_manager = self._certificate_manager or CertificateManager(self._api)
        return self._certificate_manager

    @property
    def certificate_pre_enrollment(self) -> CertificatePreEnrollment:
        self._certificate_pre_enrollment = self._certificate_pre_enrollment or CertificatePreEnrollment(self._api)
        return self._certificate_pre_enrollment

    @property
    def certificate_revocation(self) -> CertificateRevocation:
        self._certificate_revocation = self._certificate_revocation or CertificateRevocation(self._api)
        return self._certificate_revocation

    @property
    def cloud_instance_monitor(self) -> CloudInstanceMonitor:
        self._cloud_instance_monitor = self._cloud_instance_monitor or CloudInstanceMonitor(self._api)
        return self._cloud_instance_monitor

    @property
    def discovery_manager(self) -> DiscoveryManager:
        self._discovery_manager = self._discovery_manager or DiscoveryManager(self._api)
        return self._discovery_manager

    @property
    def monitor(self) -> Monitor:
        self._monitor = self._monitor or Monitor(self._api)
        return self._monitor

    @property
    def onboard_discovery_manager(self) -> OnboardDiscoveryManager:
        self._onboard_discovery_manager = self._onboard_discovery_manager or OnboardDiscoveryManager(self._api)
        return self._onboard_discovery_manager

    @property
    def reporting(self) -> Reporting:
        self._reporting = self._reporting or Reporting(self._api)
        return self._reporting

    @property
    def ssh_manager(self) -> SSHManager:
        self._ssh_manager = self._ssh_manager or SSHManager(self._api)
        return self._ssh_manager

    @property
    def trustnet_manager(self) -> TrustNetManager:
        self._trustnet_manager = self._trustnet_manager or TrustNetManager(self._api)
        return self._trustnet_manager

    @property
    def validation_manager(self) -> ValidationManager:
        self._validation_manager = self._validation_manager or ValidationManager(self._api)
        return self._validation_manager
    # endregion Component Properties

    def get(self):
        """
        Returns:
            :ref:`config_object` of the Engines root.
        """
        return self._get_config_object(object_dn=self._engines_dn)
# endregion Platform Root



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.log_channel import LogChannelAttributes


class LogSQLChannelAttributes(LogChannelAttributes, metaclass=IterableMeta):
	__config_class__ = "Log SQL Channel"
	create_sql_expression = Attribute('Create SQL Expression')
	dsn = Attribute('DSN')
	dsn_vault_id = Attribute('DSN Vault Id', min_version='18.1')
	database = Attribute('Database')
	expiration = Attribute('Expiration')
	expire_sql_expression = Attribute('Expire SQL Expression')
	last_run = Attribute('Last Run')
	max_log_age_days = Attribute('Max Log Age Days', min_version='16.2')
	table_name = Attribute('Table Name')
	timeout = Attribute('Timeout')
	view_dsn = Attribute('View DSN')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.client_work_base import ClientWorkBaseAttributes


class ClientAgentSSHDiscoveryWorkAttributes(ClientWorkBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Client Agent SSH Discovery Work"
	clear_cache_timestamp = Attribute('Clear Cache Timestamp', min_version='20.1')
	days_of_month = Attribute('Days Of Month', min_version='15.1')
	days_of_week = Attribute('Days Of Week', min_version='15.1')
	exclude_remote_mount_points = Attribute('Exclude Remote Mount Points', min_version='15.1')
	interval = Attribute('Interval', min_version='15.1')
	log_threshold = Attribute('Log Threshold', min_version='15.1')
	max_filesize = Attribute('Max Filesize', min_version='15.1')
	minimize_resource_use = Attribute('Minimize Resource Use', min_version='15.1')
	new_device_object_landing = Attribute('New Device Object Landing', min_version='15.1')
	ssh_scanner_service_path = Attribute('SSH Scanner Service Path', min_version='15.1')
	ssh_scanner_user_path = Attribute('SSH Scanner User Path', min_version='15.1')
	schedule_type = Attribute('Schedule Type', min_version='15.1')
	server_path_defaults_disabled = Attribute('Server Path Defaults Disabled', min_version='15.1')
	start_time = Attribute('Start Time', min_version='15.1')
	user_path_defaults_disabled = Attribute('User Path Defaults Disabled', min_version='15.1')



from typing import Union
from pytpp.tools.logger import logger, LogTags
from pytpp.api.session import Session
from pytpp.properties.oauth import Scope
from pytpp.api.websdk.endpoints.authorize import _Authorize
from pytpp.api.websdk.endpoints.certificates import _Certificates
from pytpp.api.websdk.endpoints.client import _Client
from pytpp.api.websdk.endpoints.codesign import _Codesign
from pytpp.api.websdk.endpoints.config import _Config
from pytpp.api.websdk.endpoints.config_schema import _ConfigSchema
from pytpp.api.websdk.endpoints.credentials import _Credentials
from pytpp.api.websdk.endpoints.crypto import _Crypto
from pytpp.api.websdk.endpoints.discovery import _Discovery
from pytpp.api.websdk.endpoints.flow import _Flow
from pytpp.api.websdk.endpoints.hsm_api import _HSMAPI
from pytpp.api.websdk.endpoints.identity import _Identity
from pytpp.api.websdk.endpoints.log import _Log
from pytpp.api.websdk.endpoints.metadata import _Metadata
from pytpp.api.websdk.endpoints.permissions import _Permissions
from pytpp.api.websdk.endpoints.pki import _PKI
from pytpp.api.websdk.endpoints.preferences import _Preferences
from pytpp.api.websdk.endpoints.processing_engines import _ProcessingEngines
from pytpp.api.websdk.endpoints.revoke import _Revoke
from pytpp.api.websdk.endpoints.secret_store import _SecretStore
from pytpp.api.websdk.endpoints.ssh import _SSH
from pytpp.api.websdk.endpoints.ssh_certificates import _SSHCertificates
from pytpp.api.websdk.endpoints.stats import _Stats
from pytpp.api.websdk.endpoints.system_status import _SystemStatus
from pytpp.api.websdk.endpoints.teams import _Teams
from pytpp.api.websdk.endpoints.workflow import _Workflow
from pytpp.api.websdk.endpoints.x509_certificate_store import _X509CertificateStore


class WebSDK:
    """
    Initializes a WebSDK session via username/password. Certificate authentication is not
    currently supported. Re-authentication occurs automatically when the API Key
    becomes invalidated. When initialized, all endpoints are also initialized.
    """
    @logger.wrap_func(log_tag=LogTags.api, mask_input_regexes=['password', 'token'])
    def __init__(self, host: str, username: str, password: str, token: str = None, application_id: str = None,
                 scope: Union[Scope, str] = None, refresh_token: str = None, proxies: dict = None,
                 certificate_path: str = None, key_file_path: str = None, verify_ssl: bool = False):
        """
        Authenticates the given user to WebSDK. The only supported method for authentication at this time
        is with a username and password. Either an OAuth bearer token can be obtained, which requires
        both an Application ID and scope to be supplied, or the X-Venafi-API-Key can be obtained, which has
        been deprecated since TPP version 20.1.

        Using the OAuth authentication method requires an API Application Integration to be created that
        defines the maximum possible scope and the users/groups that are authorized to use that scope. That
        can be accomplished through Aperture. The ``scope`` parameter simply has to define the subset of
        allowed scopes defined by that application.

        Args:
            host: Hostname or IP Address of TPP
            username: Username
            password: Password
            token: OAuth Access Bearer Token
            application_id: Application ID of the OAuth API Application Integration. Must supply ``scope``.
            scope: Scope of the OAuth API Application Integration to be used. Must supply ``application_id``.
            proxies: An OrderedDict used by the python Requests library.
            certificate_path: Absolute path to the public certificate file.
            key_file_path: Absolute path to the private key file.
            verify_ssl: If ``True``, verify the SSL certificate of the target endpoints.
        """
        # region Instance Variables
        self._host = host
        self._username = username
        self._password = password
        self._application_id = application_id
        self._oauth = None
        self._token = None
        self._proxies = proxies
        self._certificate_path = certificate_path
        self._key_file_path = key_file_path
        self._verify_ssl = verify_ssl

        # This is used by the endpoints to avoid redundancy.
        self._base_url = f'https://{host}/vedsdk'
        # endregion Instance Variables

        # region Authentication
        # This is used by the endpoints to authorize the API writes.
        self._session = Session(
            headers={'Content-Type': 'application/json'}, proxies=self._proxies,
            certificate_path=certificate_path, key_file_path=key_file_path,
            verify_ssl=verify_ssl
        )

        # Authorize the WebSDK session and store the API token.
        self.Authorize = _Authorize(self)

        if not token:
            if application_id:
                if not scope:
                    raise ValueError(f'OAuth authentication requires both an Application ID and a scope be. '
                                     f'The scope was not defined.')
                elif isinstance(scope, Scope):
                    scope = scope.to_string()

                if certificate_path and key_file_path:
                    oauth = self.Authorize.Certificate.post(
                        client_id=application_id,
                        scope=scope
                    )
                else:
                    oauth = self.Authorize.OAuth.post(
                        client_id=application_id,
                        username=username,
                        password=password,
                        scope=scope
                    )
                self._oauth = oauth
                token = oauth.access_token
            else:
                # Deprecated since TPP Version 20.1
                token = self.Authorize.post(username=username, password=password).token
        elif refresh_token:
            oauth = self.Authorize.Token.post(
                client_id=application_id,
                refresh_token=refresh_token
            )
            self._oauth = oauth
            token = oauth.access_token

        if token.endswith('=='):
            self._token = f'Bearer {token}'
            authorization_header = {
                'Authorization': self._token
            }
        else:
            self._token = token
            authorization_header = {
                'X-Venafi-API-Key': self._token
            }
        self._session.update_headers(authorization_header)
        # endregion Authentication

        # region Initialize All WebSDK Endpoints
        # Initialize the rest of the endpoints with self, which contains the base url,
        # the authorization token, and the re-authentication method.
        self.Certificates = _Certificates(self)
        self.Client = _Client(self)
        self.Codesign = _Codesign(self)
        self.Config = _Config(self)
        self.ConfigSchema = _ConfigSchema(self)
        self.Credentials = _Credentials(self)
        self.Crypto = _Crypto(self)
        self.Discovery = _Discovery(self)
        self.Flow = _Flow(self)
        self.HSMAPI = _HSMAPI(self)
        self.Identity = _Identity(self)
        self.Log = _Log(self)
        self.Metadata = _Metadata(self)
        self.Permissions = _Permissions(self)
        self.PKI = _PKI(self)
        self.Preferences = _Preferences(self)
        self.ProcessingEngines = _ProcessingEngines(self)
        self.Revoke = _Revoke(self)
        self.SecretStore = _SecretStore(self)
        self.SSH = _SSH(self)
        self.SSHCertificates = _SSHCertificates(self)
        self.Stats = _Stats(self)
        self.SystemStatus = _SystemStatus(self)
        self.Teams = _Teams(self)
        self.Workflow = _Workflow(self)
        self.X509CertificateStore = _X509CertificateStore(self)
        # endregion Initialize All WebSDK Endpoints

    def re_authenticate(self):
        """
        Performs a re-authentication using the same parameters used to authorize initially.
        """
        if self._oauth is not None:
            self.__init__(
                host=self._host,
                username=self._username,
                password=self._password,
                certificate_path=self._certificate_path,
                key_file_path=self._key_file_path,
                verify_ssl=self._verify_ssl,
                proxies=self._proxies,
                application_id=self._application_id,
                scope=self._oauth.scope,
                refresh_token=self._oauth.refresh_token,
            )
        else:
            self.__init__(
                host=self._host,
                username=self._username,
                password=self._password,
                certificate_path=self._certificate_path,
                key_file_path=self._key_file_path,
                verify_ssl=self._verify_ssl,
                proxies=self._proxies
            )



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_base import ApplicationBaseAttributes


class JuniperSASAttributes(ApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "Juniper SAS"
	external_port = Attribute('External Port')
	file_validation_disabled = Attribute('File Validation Disabled')
	internal_port = Attribute('Internal Port')
	network_validation_disabled = Attribute('Network Validation Disabled')
	reassign_ports = Attribute('Reassign Ports')
	vlan_port = Attribute('Vlan Port')



from pytpp.api.api_base import API, APIResponse, api_response_property


class _HSMAPI:
    def __init__(self, api_obj):
        self.Sign = self._Sign(api_obj=api_obj)
        self.SignJWT = self._SignJWT(api_obj=api_obj)
        self.GetGPGPublicKey = self._GetGPGPublicKey(api_obj=api_obj)

    class _Sign(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='API/Sign')
            self._url = self._url.replace('vedsdk', 'vedhsm')

        def post(self, client_info: dict, data: str, key_context: str, key_id: int, mechanism: int,
                 process_info: dict, client_mechanism: str = None, justification: str = None,
                 key_context_to_wrap: int = None, key_id_to_wrap: int = None, parameter: dict = None,
                 password: str = None, username: str = None, verify_data: bool = None,
                 wrapping_key_id: int = None):
            body = {
                'ClientInfo': client_info,
                'ClientMechanism': client_mechanism,
                'Data': data,
                'Justification': justification,
                'KeyContext': key_context,
                'KeyContextToWrap': key_context_to_wrap,
                'KeyId': key_id,
                'KeyIdToWrap': key_id_to_wrap,
                'Mechanism': mechanism,
                'Parameter': parameter,
                'Password': password,
                'ProcessInfo': process_info,
                'Username': username,
                'VerifyData': verify_data,
                'WrappingKeyId': wrapping_key_id
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result_data(self) -> str:
                    return self._from_json(key='ResultData')

                @property
                @api_response_property()
                def success(self) -> bool:
                    return self._from_json(key='Success')

                @property
                @api_response_property()
                def try_later(self) -> bool:
                    return self._from_json(key='TryLater')

            return _Response(response=self._post(data=body))

    class _SignJWT(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='API/SignJWT')
            self._url = self._url.replace('vedsdk', 'vedhsm')

        def post(self, client_info: dict, process_info: dict, key_id: str,
                 header: str, payload: str):
            body = {
                'ClientInfo': client_info,
                'ProcessInfo': process_info,
                'KeyId': key_id,
                'Header': header,
                'Payload': payload
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def result_data(self) -> str:
                    return self._from_json(key='ResultData')

                @property
                @api_response_property()
                def success(self) -> bool:
                    return self._from_json(key='Success')

            return _Response(response=self._post(data=body))

    class _GetGPGPublicKey(API):
        def __init__(self, api_obj):
            super().__init__(api_obj=api_obj, url='API/GetGPGPublicKey')
            self._url = self._url.replace('vedsdk', 'vedhsm')

        def post(self, key_id: str, key_context: str = None):
            body = {
                'KeyId':key_id,
                'KeyContext': key_context
            }

            class _Response(APIResponse):
                def __init__(self, response):
                    super().__init__(response=response)

                @property
                @api_response_property()
                def fingerprint(self) -> str:
                    return self._from_json(key='Fingerprint')

                @property
                @api_response_property()
                def location(self) -> str:
                    return self._from_json(key='Location')

                @property
                @api_response_property()
                def public_key(self) -> str:
                    return self._from_json(key='PublicKey')

                @property
                @api_response_property()
                def success(self) -> bool:
                    return self._from_json(key='Success')

            return _Response(response=self._post(data=body))




from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.client_work_base import ClientWorkBaseAttributes


class SSHWorkAttributes(ClientWorkBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "SSH Work"
	days_of_month = Attribute('Days Of Month')
	days_of_week = Attribute('Days Of Week')
	exclude_remote_mount_points = Attribute('Exclude Remote Mount Points')
	interval = Attribute('Interval')
	max_filesize = Attribute('Max Filesize')
	new_device_object_landing = Attribute('New Device Object Landing')
	remediation_interval = Attribute('Remediation Interval')
	remediation_schedule_type = Attribute('Remediation Schedule Type')
	remediation_start_time = Attribute('Remediation Start Time')
	ssh_scanner_service_path = Attribute('SSH Scanner Service Path')
	ssh_scanner_user_path = Attribute('SSH Scanner User Path')
	schedule_type = Attribute('Schedule Type')
	server_path_defaults_disabled = Attribute('Server Path Defaults Disabled')
	start_time = Attribute('Start Time')
	user_path_defaults_disabled = Attribute('User Path Defaults Disabled')



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.adaptable_flow_action_base import AdaptableFlowActionBaseAttributes


class SSHCertificateAdaptableFlowActionAttributes(AdaptableFlowActionBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "SSH Certificate Adaptable Flow Action"
	allow_to_modify_certificate_request = Attribute('Allow To Modify Certificate Request', min_version='21.4')
	certificate_credential = Attribute('Certificate Credential', min_version='21.4')
	ssh_certificate_adaptable_flow_action_text_field_1 = Attribute('SSH Certificate Adaptable Flow Action Text Field 1', min_version='21.4')
	ssh_certificate_adaptable_flow_action_text_field_10 = Attribute('SSH Certificate Adaptable Flow Action Text Field 10', min_version='21.4')
	ssh_certificate_adaptable_flow_action_text_field_11 = Attribute('SSH Certificate Adaptable Flow Action Text Field 11', min_version='21.4')
	ssh_certificate_adaptable_flow_action_text_field_12 = Attribute('SSH Certificate Adaptable Flow Action Text Field 12', min_version='21.4')
	ssh_certificate_adaptable_flow_action_text_field_2 = Attribute('SSH Certificate Adaptable Flow Action Text Field 2', min_version='21.4')
	ssh_certificate_adaptable_flow_action_text_field_3 = Attribute('SSH Certificate Adaptable Flow Action Text Field 3', min_version='21.4')
	ssh_certificate_adaptable_flow_action_text_field_4 = Attribute('SSH Certificate Adaptable Flow Action Text Field 4', min_version='21.4')
	ssh_certificate_adaptable_flow_action_text_field_5 = Attribute('SSH Certificate Adaptable Flow Action Text Field 5', min_version='21.4')
	ssh_certificate_adaptable_flow_action_text_field_6 = Attribute('SSH Certificate Adaptable Flow Action Text Field 6', min_version='21.4')
	ssh_certificate_adaptable_flow_action_text_field_7 = Attribute('SSH Certificate Adaptable Flow Action Text Field 7', min_version='21.4')
	ssh_certificate_adaptable_flow_action_text_field_8 = Attribute('SSH Certificate Adaptable Flow Action Text Field 8', min_version='21.4')
	ssh_certificate_adaptable_flow_action_text_field_9 = Attribute('SSH Certificate Adaptable Flow Action Text Field 9', min_version='21.4')
	secondary_credential = Attribute('Secondary Credential', min_version='21.4')
	service_address = Attribute('Service Address', min_version='21.4')
	username_credential = Attribute('Username Credential', min_version='21.4')
	validate_certificate_request_before_issuance = Attribute('Validate Certificate Request Before Issuance', min_version='21.4')



from dataclasses import dataclass
from typing import List, Dict, Union
from pytpp.tools.vtypes import Config, CustomFields
from pytpp.features.bases.feature_base import FeatureBase, feature
from pytpp.features.definitions.exceptions import InvalidResultCode

@dataclass
class EffectiveValues:
    locked: bool
    policy_dn: str
    values: list


@dataclass
class PolicyValues:
    locked: bool
    values: list


@feature('Custom Field')
class CustomField(FeatureBase):
    def __init__(self, api):
        super().__init__(api)
        self._metadata_root_dn = r'\VED\Metadata Root'

    @staticmethod
    def _guid_data_list(guid_data_dict: dict):
        return [
            {"ItemGuid": key, "List": value}
            for key, value in guid_data_dict.items()
        ]

    @staticmethod
    def _validate_result_code(result):
        if result.code != 0:
            raise InvalidResultCode(code=result.code, code_description=result.metadata_result)

    def create(self, label: str, name: str, classes: list, data_type: int, allowed_characters: List[str] = None,
               allowed_values: List[str] = None, category: str = None, date_only: bool = None, default_values: List = None,
               display_after: str = None, error_message: str = None, help_text: str = None, localization_table: str = None,
               localized_help: Union[List[str], Dict] = None, localized_label: str = None, localized_set: str = None,
               mandatory: bool = None, mask: str = None, maximum_length: int = None, minimum_length: int = None,
               policyable: bool = None, regular_expression: str = None, render_hidden: bool = None,
               render_read_only: bool = None, single: bool = None, time_only: bool = None,
               get_if_already_exists: bool = True):
        """
        Creates a Custom Field (or Metadata Item). Custom fields appear in the UI unless ``render_hidden = True``. If
        the custom field should be implemented on a folder policy, then set ``policyable=True``, otherwise the custom
        field cannot be enforced on a policy, but will instead apply to all objects that belong to one of the ``classes``.
        If it is configured to be policyable, the ``classes`` still apply. For example, if one of the classes is set to
        "X509 Certificate", then all objects that are a "X509 Certificate" will inherit the custom field. If it is set
        to be policyable and is locked to a specific value, only "X509 Certificate" objects under the policy folder
        will have that value locked.

        Args:
            label: The label text for the custom field.
            name: The name of the custom field object.
            classes: The classes that this field applies to.
            data_type: The field type as an integer.
            allowed_characters: A set of allowable alphabetic, numeric, or symbols.
            allowed_values: An array of allowable values.
            category: Absolute path to the item category.
            date_only: If ``True``, a calendar date is required.
            default_values: List of default values.
            display_after: Appears after a list of Custom Field values.
            error_message: The error message when an error is encountered.
            help_text: The tool tip for the UI.
            localization_table: A list or dictionary of strings from another language.
            localized_help: The localized help text that describes the Metadata Custom Field.
            localized_label: A localized label text describing the Metadata Custom Field.
            localized_set: A list of localization settings.
            mandatory: If ``True``, this field is required.
            mask: The set of characters to use to hide user input in the UI.
            maximum_length: Maximum number of allowed characters.
            minimum_length: Minimum number of allowed characters.
            policyable: If ``True``, the custom field can be enforced on a policy folder.
            regular_expression: The regular expression to control user input.
            render_hidden: If ``True``, hides the field from the UI.
            render_read_only: If ``False``, the user can change the default values (if not locked by policy).
            single: If ``False``, the user can select/input multiple items.
            time_only: If ``True``, time of day is required.
            get_if_already_exists: If the objects already exists, just return it as is.

        Returns:
            The custom field :class:`~.dataclasses.metadata.Item` data.
        """
        item = {
            'AllowedCharacters': allowed_characters,
            'AllowedValues': allowed_values,
            'Category': category,
            'Classes': classes,
            'DateOnly': date_only,
            'DefaultValues': default_values,
            'DisplayAfter': display_after,
            'ErrorMessage': error_message,
            'Help': help_text,
            'Label': label,
            'LocalizationTable': localization_table,
            'LocalizedHelp': localized_help,
            'LocalizedLabel': localized_label,
            'LocalizedSet': localized_set,
            'Mandatory': mandatory,
            'Name': name,
            'Mask': mask,
            'MaximumLength': maximum_length,
            'MinimumLength': minimum_length,
            'Policyable': policyable,
            'RegularExpression': regular_expression,
            'RenderHidden': render_hidden,
            'RenderReadOnly': render_read_only,
            'Single': single,
            'TimeOnly': time_only,
            'Type': data_type
        }

        response = self._api.websdk.Metadata.DefineItem.post(item=item)
        if response.result.code == 9 and get_if_already_exists:
            return self.get_item_details(custom_field=name)
        self._validate_result_code(response.result)
        return response.item

    def delete(self, custom_field: Union['Config.Object', 'CustomFields.Item', str], remove_data: bool = True):
        """
        Deletes a custom field and all instances of it, including policy settings. If ``remove_data = False``, then
        an exception will be raised if there is existing data for the custom field on any object.

        Args:
            custom_field: :ref:`config_object` or name of the custom field.
            remove_data: If ``True``, then deletes the custom field regardless of data existing on the custom field
                on any object that implements it.
        """
        custom_field_guid = self._get_guid(custom_field, parent_dn=self._metadata_root_dn)
        response = self._api.websdk.Metadata.UndefineItem.post(
            item_guid=custom_field_guid,
            remove_data=remove_data
        )
        self._validate_result_code(response.result)

    def get(self, name: str = None, raise_error_if_not_exists: bool = True):
        """
        Args:
            name: The name of the custom field object.
            raise_error_if_not_exists: Raise an exception if the custom field does not exist.

        Returns:
            :ref:`config_object` of the custom field object.
        """
        return self._get_config_object(
            object_dn=f'{self._metadata_root_dn}\\{name}',
            raise_error_if_not_exists=raise_error_if_not_exists
        )

    def get_item_details(self, custom_field: Union['Config.Object', 'CustomFields.Item', str] = None):
        """
        Args:
            custom_field: :ref:`config_object` or name of the custom field object.

        Returns:
            The custom field :class:`~.dataclasses.metadata.Item` data.
        """
        custom_field_dn = self._get_dn(custom_field, parent_dn=self._metadata_root_dn)
        response = self._api.websdk.Metadata.LoadItem.post(dn=custom_field_dn)
        self._validate_result_code(result=response.result)
        return response.item

    def list(self):
        """
        Lists all custom fields registered in TPP.

        Returns:
            List of each custom field :class:`~.dataclasses.metadata.Item` data.
        """
        response = self._api.websdk.Metadata.Items.get()
        self._validate_result_code(response.result)
        return response.items

    def read(self, obj: 'Union[Config.Object, str]', custom_field: Union['Config.Object', 'CustomFields.Item', str]) -> EffectiveValues:
        """
        Reads the effective value(s) of a custom field (which includes for policy values). Value(s) may be None.

        Args:
            obj: :ref:`config_object` or :ref:`dn` of the object having the custom field.
            custom_field: :ref:`config_object` or name of the custom field..

        Returns:
            A *EffectiveValues* object with these properties

            * **locked** *bool* - If ``True``, the ``values`` are locked by the policy.
            * **policy_dn** *str* - :ref:`dn` of the policy folder locking the values.
            * **values** *List[str]* - List of values.
        """
        obj_dn = self._get_dn(obj)
        custom_field_guid = self._get_guid(custom_field, parent_dn=self._metadata_root_dn)
        response = self._api.websdk.Metadata.ReadEffectiveValues.post(
            dn=obj_dn,
            item_guid=custom_field_guid
        )
        self._validate_result_code(response.result)
        return EffectiveValues(locked=response.locked, values=response.values, policy_dn=response.policy_dn)

    def read_policy(self, folder: 'Union[Config.Object, str]', custom_field: Union['Config.Object', 'CustomFields.Item', str],
                    class_name: str) -> PolicyValues:
        """
        Reads the effective value(s) of a custom field set on a policy. Value(s) may be None.

        Args:
            folder: :ref:`config_object` or :ref:`dn` of the folder.
            custom_field: :ref:`config_object` or name of the custom field. 
            class_name: Object class.

        Returns:
            A *PolicyValues* object with these properties

            * **locked** *bool* - If ``True``, the ``values`` are locked by the policy.
            * **values** *List[str]* - List of values.
        """
        folder_dn = self._get_dn(folder)
        custom_field_guid = self._get_guid(custom_field, parent_dn=self._metadata_root_dn)
        response = self._api.websdk.Metadata.ReadPolicy.post(
            dn=folder_dn,
            item_guid=custom_field_guid,
            obj_type=str(class_name)
        )
        self._validate_result_code(response.result)
        return PolicyValues(locked=response.locked, values=response.values)

    def update(self, custom_field: Union['Config.Object', 'CustomFields.Item', str], allowed_characters: List[str] = None,
               allowed_values: List[str] = None, category: str = None, classes: list = None, data_type: int = None,
               date_only: bool = None, default_values: str = None, display_after: str = None, error_message: str = None,
               help_text: str = None, label: str = None, localization_table: str = None, localized_help: str = None,
               localized_label: str = None, localized_set: str = None, mandatory: bool = None, mask: str = None,
               maximum_length: int = None, minimum_length: int = None, name: str = None, policyable: bool = None,
               regular_expression: str = None, render_hidden: bool = None, render_read_only: bool = None, single: bool = None,
               time_only: bool = None):
        """
        Updates a custom field's properties. If a parameter is ``None`` it does not affect the current setting of that
        property. Property values are not reset to their defaults.

        Args:
            custom_field: :ref:`config_object` or name of the metadata object.
            label: The label text for the custom field.
            name: The name of the custom field object.
            classes: The classes that this field applies to.
            data_type: The field type as an integer.
            allowed_characters: A set of allowable alphabetic, numeric, or symbols.
            allowed_values: An array of allowable values.
            category: Absolute path to the item category.
            date_only: If ``True``, a calendar date is required.
            default_values: List of default values.
            display_after: Appears after a list of Custom Field values.
            error_message: The error message when an error is encountered.
            help_text: The tool tip for the UI.
            localization_table: A list or dictionary of strings from another language.
            localized_help: The localized help text that describes the Metadata Custom Field.
            localized_label: A localized label text describing the Metadata Custom Field.
            localized_set: A list of localization settings.
            mandatory: If ``True``, this field is required.
            mask: The set of characters to use to hide user input in the UI.
            maximum_length: Maximum number of allowed characters.
            minimum_length: Minimum number of allowed characters.
            policyable: If ``True``, the custom field can be enforced on a policy folder.
            regular_expression: The regular expression to control user input.
            render_hidden: If ``True``, hides the field from the UI.
            render_read_only: If ``False``, the user can change the default values (if not locked by policy).
            single: If ``False``, the user can select/input multiple items.
            time_only: If ``True``, time of day is required.

        Returns:
            The custom field :class:`~.dataclasses.metadata.Item` data.
        """
        custom_field_dn = self._get_dn(custom_field, parent_dn=self._metadata_root_dn)
        listify = lambda x: [x] if x and not isinstance(x, list) else x

        item = {
            'AllowedCharacters': listify(allowed_characters),
            'AllowedValues': listify(allowed_values),
            'Category': listify(category),
            'Classes': listify(classes),
            'DateOnly': listify(date_only),
            'DefaultValues': listify(default_values),
            'DisplayAfter': listify(display_after),
            'ErrorMessage': listify(error_message),
            'Help': listify(help_text),
            'Label': label,
            'LocalizationTable': listify(localization_table),
            'LocalizedHelp': listify(localized_help),
            'LocalizedLabel': listify(localized_label),
            'LocalizedSet': listify(localized_set),
            'Mandatory': listify(mandatory),
            'Name': listify(name),
            'Mask': listify(mask),
            'MaximumLength': listify(maximum_length),
            'MinimumLength': listify(minimum_length),
            'Policyable': listify(policyable),
            'RegularExpression': listify(regular_expression),
            'RenderHidden': listify(render_hidden),
            'RenderReadOnly': listify(render_read_only),
            'Single': listify(single),
            'TimeOnly': listify(time_only),
            'Type': listify(data_type)
        }
        response = self._api.websdk.Metadata.UpdateItem.post(
            update={
                "DN": custom_field_dn,
                "Data": self._name_value_list(attributes=item, keep_list_values=True)
            }
        )
        self._validate_result_code(response.result)
        return self.get_item_details(custom_field=custom_field)

    def write(self, obj: 'Union[Config.Object, str]', custom_field: Union['Config.Object', 'CustomFields.Item', str], values: List,
              keep_existing: bool = True):
        """
        Writes a set of values to a custom field on the specified ``obj``. If ``keep_existing = False``,
        then all custom fields related to the ``obj`` are reset to their default values.

        Args:
            obj: :ref:`config_object` or :ref:`dn` of the object with the custom field.
            custom_field: :ref:`config_object` or name of the custom field. object.
            values: List of values to write to ``object_dn``.
            keep_existing: If ``True``, all other custom fields remain unaffected.
        """
        obj_dn = self._get_dn(obj)
        custom_field_guid = self._get_guid(custom_field)
        response = self._api.websdk.Metadata.Set.post(
            dn=obj_dn,
            guid_data=self._guid_data_list({custom_field_guid: values}),
            keep_existing=keep_existing
        )
        self._validate_result_code(result=response.result)

    def write_policy(self, folder: 'Union[Config.Object, str]', custom_field: Union['Config.Object', 'CustomFields.Item', str], class_name: str,
                     values: List, locked: bool = False):
        """
        Writes a set of values to a custom field on the specified ``folder``. The custom field MUST be policyable.
        To ensure, use :meth:``get`` to get the current settings and verify that ``policyable=True``. If not, use
        :meth:``update`` to update the custom field with ``policyable=True``.

        Args:
            folder: :ref:`config_object` or :ref:`dn` of the folder.
            custom_field: :ref:`config_object` or name of the custom field. object.
            class_name: Name of the object class.
            values: List of values to write to ``obj``.
            locked: If ``True``, all subordinate objects inherit and cannot modify the ``values``.
        """
        folder_dn = self._get_dn(folder)
        custom_field_guid = self._get_guid(custom_field)
        response = self._api.websdk.Metadata.SetPolicy.post(
            dn=folder_dn,
            config_class=str(class_name),
            guid_data=self._guid_data_list({custom_field_guid: values}),
            locked=locked
        )
        self._validate_result_code(result=response.result)



from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.top import TopAttributes


class TreeRootAttributes(TopAttributes, metaclass=IterableMeta):
	__config_class__ = "Tree Root"
	company_name = Attribute('Company Name', min_version='17.3')
	migration_task = Attribute('Migration Task')
	pendo_eula_version = Attribute('Pendo EULA Version', min_version='19.2')
	pendo_optional_data_collection = Attribute('Pendo Optional Data Collection', min_version='19.2')
	schema_version = Attribute('Schema Version')
	usage_tracking = Attribute('Usage Tracking', min_version='19.2')
	use_company_name_for_analytics = Attribute('Use Company Name for Analytics', min_version='17.3')
	version = Attribute('Version')



class SSHKeydataObject:
    active_from = 'ActiveFrom'
    algorithm = 'Algorithm'
    allowed_source_restriction = 'AllowedSourceRestriction'
    approver = 'Approver'
    denied_source_restriction = 'DeniedSourceRestriction'
    device_guid = 'DeviceGuid'
    filepath = 'Filepath'
    forced_command = 'ForcedCommand'
    format = 'Format'
    is_encrypted = 'IsEncrypted'
    key_id = 'KeyId'
    keysetid = 'Keysetid'
    last_used = 'Last Used'
    length = 'Length'
    notes = 'Notes'
    options = 'Options '
    process_error = 'ProcessError'
    process_status = 'ProcessStatus'
    reason = 'Reason'
    type = 'Type'
    username = 'Username'
    violation_status = 'ViolationStatus'




from pytpp.attributes._helper import IterableMeta, Attribute
from pytpp.attributes.application_base import ApplicationBaseAttributes


class PKCS11Attributes(ApplicationBaseAttributes, metaclass=IterableMeta):
	__config_class__ = "PKCS11"
	file_validation_disabled = Attribute('File Validation Disabled')
	hsm_cblob = Attribute('HSM:CBlob')
	hsm_cka_label_format = Attribute('HSM:CKA LABEL Format')
	hsm_csr_subject_dn = Attribute('HSM:CSR Subject DN')
	hsm_certificate_directory = Attribute('HSM:Certificate Directory')
	hsm_client_tool_path = Attribute('HSM:Client Tool Path')
	hsm_cryptoki_file = Attribute('HSM:Cryptoki File')
	hsm_embed_sans_in_csr = Attribute('HSM:Embed SANs in CSR')
	hsm_ic_cka_label_format = Attribute('HSM:IC CKA LABEL Format')
	hsm_import_certificate = Attribute('HSM:Import Certificate')
	hsm_issued_id = Attribute('HSM:Issued ID')
	hsm_issued_label = Attribute('HSM:Issued Label')
	hsm_issued_usecase = Attribute('HSM:Issued Usecase')
	hsm_kpblob = Attribute('HSM:KPBlob')
	hsm_last_issued_cblob = Attribute('HSM:Last Issued CBlob')
	hsm_last_issued_id = Attribute('HSM:Last Issued ID')
	hsm_last_issued_kpblob = Attribute('HSM:Last Issued KPBlob')
	hsm_last_issued_label = Attribute('HSM:Last Issued Label')
	hsm_last_issued_usecase = Attribute('HSM:Last Issued Usecase')
	hsm_openssl_config_file = Attribute('HSM:Openssl Config File')
	hsm_openssl_path = Attribute('HSM:Openssl Path')
	hsm_openssl_type = Attribute('HSM:Openssl Type')
	hsm_pkcs11attributes = Attribute('HSM:PKCS11Attributes')
	hsm_protection_type = Attribute('HSM:Protection Type')
	hsm_requested_cka_label = Attribute('HSM:Requested CKA LABEL')
	hsm_requested_ecdh = Attribute('HSM:Requested ECDH')
	hsm_requested_usecase = Attribute('HSM:Requested Usecase')
	hsm_reverse_subject_dn = Attribute('HSM:Reverse Subject DN')
	hsm_tmp_issued_cblob = Attribute('HSM:TMP Issued CBlob')
	hsm_tmp_issued_id = Attribute('HSM:TMP Issued ID')
	hsm_tmp_issued_kpblob = Attribute('HSM:TMP Issued KPBlob')
	hsm_token_label = Attribute('HSM:Token Label')
	hsm_token_password = Attribute('HSM:Token Password')
	network_validation_disabled = Attribute('Network Validation Disabled')



from packaging.version import Version
from typing import List, Generator


class Attribute(str):
    def __init__(self, name: str, min_version: str = None):
        self.name = name
        self.min_version = min_version
        if min_version is not None:
            self.min_version = Version('.'.join(self.min_version.split('.', maxsplit=2)[:2]))

    def __new__(cls, name: str, *args, **kwargs):
        return str().__new__(cls, name)


class IterableMeta(type):
    __config_class__ = None

    def __iter__(self) -> Generator[Attribute, None, None]:
        for item in dir(self):
            if not item.startswith('_'):
                attr = getattr(self, item)
                if not callable(attr):
                    yield attr

    def list(cls) -> List[Attribute]:
        return list(iter(cls))  # type: List[Attribute]

    def __repr__(self):
        return self.__config_class__

    def __str__(self):
        return self.__config_class__
